import os
import time

# noinspection PyPackageRequirements
from PyOBEX import headers, requests, responses
# noinspection PyPackageRequirements
from PyOBEX.server import Server

from platypush.backend import Backend
from platypush.message.event.bluetooth import BluetoothDeviceConnectedEvent, BluetoothFileReceivedEvent, \
    BluetoothDeviceDisconnectedEvent, BluetoothFilePutRequestEvent


class BluetoothBackend(Backend, Server):
    _sleep_on_error = 10.0

    def __init__(self, address: str = '', port: int = None, directory: str = None, whitelisted_addresses=None,
                 **kwargs):
        Backend.__init__(self, **kwargs)
        Server.__init__(self, address=address)
        self.port = port
        self.directory = os.path.join(os.path.expanduser(directory))
        self.whitelisted_addresses = whitelisted_addresses or []
        self._sock = None

    def run(self):
        self.logger.info('Starting bluetooth service [address={}] [port={}]'.format(
            self.address, self.port))

        while not self.should_stop():
            try:
                # noinspection PyArgumentList
                self._sock = self.start_service(self.port)
                self.serve(self._sock)
            except Exception as e:
                self.logger.error('Error on bluetooth connection [address={}] [port={}]: {}'.format(
                    self.address, self.port, str(e)))
                time.sleep(self._sleep_on_error)
            finally:
                self.stop()

    def stop(self):
        if self._sock:
            self.stop_service(self._sock)
        self._sock = None

    def put(self, socket, request):
        name = ""
        body = ""

        while True:
            for header in request.header_data:
                if isinstance(header, headers.Name):
                    name = header.decode()
                    self.logger.info("Receiving {}".format(name))
                elif isinstance(header, headers.Length):
                    length = header.decode()
                    self.logger.info("Content length: {} bytes".format(length))
                elif isinstance(header, headers.Body):
                    body += header.decode()
                elif isinstance(header, headers.End_Of_Body):
                    body += header.decode()

            if request.is_final():
                break

            # Ask for more data.
            Server.send_response(self, socket, responses.Continue())

            # Get the next part of the data.
            request = self.request_handler.decode(socket)

        Server.send_response(self, socket, responses.Success())
        name = os.path.basename(name.strip("\x00"))
        path = os.path.join(self.directory, name)

        self.logger.info("Writing file {}" .format(path))
        open(path, "wb").write(body.encode())
        self.bus.post(BluetoothFileReceivedEvent(path=path))

    def process_request(self, connection, request, *address):
        if isinstance(request, requests.Connect):
            self.connect(connection, request)
            self.bus.post(BluetoothDeviceConnectedEvent(address=address[0], port=address[1]))
        elif isinstance(request, requests.Disconnect):
            self.disconnect(connection)
            self.bus.post(BluetoothDeviceDisconnectedEvent(address=address[0], port=address[1]))
        elif isinstance(request, requests.Put):
            self.bus.post(BluetoothFilePutRequestEvent(address=address[0], port=address[1]))
            self.put(connection, request)
        else:
            self._reject(connection)
            self.bus.post(BluetoothFilePutRequestEvent(address=address[0], port=address[1]))

    def accept_connection(self, address, port):
        return address in self.whitelisted_addresses


# vim:sw=4:ts=4:et:



import base64
import json

from platypush.backend import Backend
from platypush.message.event.nfc import NFCTagDetectedEvent, NFCTagRemovedEvent, NFCDeviceConnectedEvent, \
    NFCDeviceDisconnectedEvent


class NfcBackend(Backend):
    """
    Backend to detect NFC card events from a compatible reader.

    Triggers:

        * :class:`platypush.message.event.nfc.NFCDeviceConnectedEvent` when an NFC reader/writer is connected
        * :class:`platypush.message.event.nfc.NFCDeviceDisconnectedEvent` when an NFC reader/writer is disconnected
        * :class:`platypush.message.event.nfc.NFCTagDetectedEvent` when an NFC tag is detected
        * :class:`platypush.message.event.nfc.NFCTagRemovedEvent` when an NFC tag is removed

    Requires:

        * **nfcpy** >= 1.0 (``pip install 'nfcpy>=1.0'``)
        * **ndef** (``pip install ndef``)

    Run the following to check if your device is compatible with nfcpy and the right permissions are set::

        python -m nfc

    """

    def __init__(self, device='usb', *args, **kwargs):
        """
        :param device: Address or ID of the device to be opened. Examples:

            * `'usb:003:009'` opens device 9 on bus 3
            * `'usb:003'` opens the first available device on bus 3
            * `'usb'` opens the first available USB device (default)
        """

        super().__init__(*args, **kwargs)

        self.device_id = device
        self._clf = None

    def _get_clf(self):
        import nfc

        if not self._clf:
            self._clf = nfc.ContactlessFrontend()
            self._clf.open(self.device_id)
            self.bus.post(NFCDeviceConnectedEvent(reader=self._get_device_str()))
            self.logger.info('Initialized NFC reader backend on device {}'.format(self._get_device_str()))

        return self._clf

    def _get_device_str(self):
        return str(self._clf.device)

    def close(self):
        if self._clf:
            self._clf.close()
            self._clf = None
            self.bus.post(NFCDeviceDisconnectedEvent(reader=self._get_device_str()))

    @staticmethod
    def _parse_records(tag):
        import ndef
        records = []

        if not tag.ndef:
            return records

        for record in tag.ndef.records:
            r = {
                'record_type': record.type,
                'record_name': record.name,
            }

            if isinstance(record, ndef.TextRecord):
                try:
                    r = {
                        **r,
                        'type': 'json',
                        'value': json.loads(record.text),
                    }
                except ValueError:
                    r = {
                        **r,
                        'type': 'text',
                        'text': record.text,
                    }
            elif isinstance(record, ndef.UriRecord):
                r = {
                    **r,
                    'type': 'uri',
                    'uri': record.uri,
                    'iri': record.iri,
                }
            elif isinstance(record, ndef.SmartposterRecord):
                r = {
                    **r,
                    'type': 'smartposter',
                    **{attr: getattr(record, attr) for attr in ['resource', 'titles', 'title', 'action', 'icon',
                                                                'icons', 'resource_size', 'resource_type']},
                }
            elif isinstance(record, ndef.DeviceInformationRecord):
                r = {
                    **r,
                    'type': 'device_info',
                    **{attr: getattr(record, attr) for attr in ['vendor_name', 'model_name', 'unique_name',
                                                                'uuid_string', 'version_string']},
                }
            elif isinstance(record, ndef.WifiSimpleConfigRecord):
                r = {
                    **r,
                    'type': 'wifi_simple_config',
                    **{attr: record[attr] for attr in record.attribute_names()}
                }
            elif isinstance(record, ndef.WifiPeerToPeerRecord):
                r = {
                    **r,
                    'type': 'wifi_peer_to_peer',
                    **{attr: record[attr] for attr in record.attribute_names()}
                }
            elif isinstance(record, ndef.BluetoothEasyPairingRecord):
                r = {
                    **r,
                    'type': 'bluetooth_easy_pairing',
                    **{attr: getattr(record, attr) for attr in ['device_address', 'device_name', 'device_class']},
                }
            elif isinstance(record, ndef.BluetoothLowEnergyRecord):
                r = {
                    **r,
                    'type': 'bluetooth_low_energy',
                    **{attr: getattr(record, attr) for attr in ['device_address', 'device_name', 'role_capabilities',
                                                                'appearance', 'flags', 'security_manager_tk_value',
                                                                'secure_connections_confirmation_value',
                                                                'secure_connections_random_value']},
                }
            elif isinstance(record, ndef.SignatureRecord):
                r = {
                    **r,
                    'type': 'signature',
                    **{attr: getattr(record, attr) for attr in ['version', 'signature_type', 'hash_type', 'signature',
                                                                'signature_uri', 'certificate_format',
                                                                'certificate_store', 'certificate_uri',
                                                                'secure_connections_random_value']},
                }
            else:
                r = {
                    **r,
                    'type': 'binary',
                    'data': base64.encodebytes(record.data).decode(),
                }

            records.append(r)

        return records

    @staticmethod
    def _parse_id(tag):
        return ''.join([('%02X' % c) for c in tag.identifier])

    def _on_connect(self):
        def callback(tag):
            if not tag:
                return False

            tag_id = self._parse_id(tag)
            records = self._parse_records(tag)
            self.bus.post(NFCTagDetectedEvent(reader=self._get_device_str(), tag_id=tag_id, records=records))
            return True

        return callback

    def _on_release(self):
        def callback(tag):
            tag_id = self._parse_id(tag)
            self.bus.post(NFCTagRemovedEvent(reader=self._get_device_str(), tag_id=tag_id))

        return callback

    def run(self):
        super().run()

        while not self.should_stop():
            try:
                clf = self._get_clf()
                clf.connect(rdwr={
                    'on-connect': self._on_connect(),
                    'on-release': self._on_release(),
                })
            finally:
                self.close()


# vim:sw=4:ts=4:et:



import redis
import socket
import threading

from platypush.backend import Backend
from platypush.message import Message
from platypush.message.request import Request
from platypush.utils import set_thread_name


class TcpBackend(Backend):
    """
    Backend that reads messages from a configured TCP port
    """

    # Maximum length of a request to be processed
    _MAX_REQ_SIZE = 2048

    def __init__(self, port, bind_address=None, listen_queue=5, *args, **kwargs):
        """
        :param port: TCP port number
        :type port: int

        :param bind_address: Specify a bind address if you want to hook the service to a specific interface (default: listen for any connections)
        :type bind_address: str

        :param listen_queue: Maximum number of queued connections (default: 5)
        :type listen_queue: int
        """

        super().__init__(*args, **kwargs)

        self.port = port
        self.bind_address = bind_address or '0.0.0.0'
        self.listen_queue = listen_queue

    def _process_client(self, sock, address):
        def _f():
            processed_bytes = 0
            open_brackets = 0
            msg = b''
            prev_ch = None
            redis = self._get_redis()

            while True:
                if processed_bytes > self._MAX_REQ_SIZE:
                    self.logger.warning('Ignoring message longer than {} bytes from {}'
                                        .format(self._MAX_REQ_SIZE, address[0]))
                    return

                ch = sock.recv(1)
                processed_bytes += 1

                if ch == b'':
                    break

                if ch == b'{' and prev_ch != b'\\':
                    open_brackets += 1

                if not open_brackets:
                    continue

                msg += ch

                if ch == b'}' and prev_ch != b'\\':
                    open_brackets -= 1

                if not open_brackets:
                    break

                prev_ch = ch

            if msg == b'':
                return

            msg = Message.build(msg)
            self.logger.info('Received request from {}: {}'.format(msg, address[0]))
            self.on_message(msg)

            response = self.get_message_response(msg)
            self.logger.info('Processing response on the TCP backend: {}'.format(response))

            if response:
                sock.send(str(response).encode())

        def _f_wrapper():
            set_thread_name('TCPListener')
            try:
                _f()
            finally:
                sock.close()

        threading.Thread(target=_f_wrapper, name='TCPListener').start()

    def run(self):
        super().run()

        serv_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        serv_sock.bind((self.bind_address, self.port))

        self.logger.info('Initialized TCP backend on port {} with bind address {}'.
                     format(self.port, self.bind_address))

        serv_sock.listen(self.listen_queue)

        while not self.should_stop():
            (sock, address) = serv_sock.accept()
            self.logger.info('Accepted connection from client {}'.format(address[0]))
            self._process_client(sock, address)


# vim:sw=4:ts=4:et:




import json
import os
import requests

from requests.auth import HTTPBasicAuth
from typing import Optional, Union, Dict, List, Any

from platypush.message.response.camera.android import AndroidCameraStatusResponse, AndroidCameraStatusListResponse, \
    AndroidCameraPictureResponse
from platypush.plugins import Plugin, action


class AndroidIpcam:
    args = {}

    def __init__(self, name: str, host: str, port: int = 8080, username: Optional[str] = None,
                 password: Optional[str] = None, timeout: int = 10, ssl: bool = True):
        self.args = {
            'name': name,
            'host': host,
            'port': port,
            'username': username,
            'password': password,
            'timeout': timeout,
            'ssl': ssl,
        }

        self.auth = None
        if username:
            self.auth = HTTPBasicAuth(self.username, self.password)

    def __getattr__(self, item):
        if item in self.args:
            return super().__getattribute__('args').get(item)
        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if key not in self.args:
            super().__setattr__(key, value)
        else:
            self.args[key] = value

    def __str__(self):
        return json.dumps(getattr(self, 'args') or {})

    @property
    def base_url(self) -> str:
        return 'http{ssl}://{host}:{port}/'.format(
            ssl=('s' if self.ssl else ''), host=self.host, port=self.port)

    @property
    def stream_url(self) -> str:
        return self.base_url + 'video'

    @property
    def audio_url(self) -> str:
        return self.base_url + 'audio.wav'

    @property
    def image_url(self) -> str:
        return self.base_url + 'photo.jpg'


class CameraAndroidIpcamPlugin(Plugin):
    """
    Plugin to control remote Android cameras over
    `IPCam <https://play.google.com/store/apps/details?id=com.pas.webcam>`_.
    """

    def __init__(self,
                 host: Optional[str] = None,
                 port: Optional[int] = 8080,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 timeout: int = 10,
                 ssl: bool = True,
                 cameras: Optional[Dict[str, Dict[str, Any]]] = None,
                 **kwargs):
        """
        :param host: Camera host name or address
        :param port: Camera port
        :param username: Camera username, if set
        :param password: Camera password, if set
        :param timeout: Connection timeout
        :param ssl: Use HTTPS instead of HTTP
        :param cameras: Alternatively, you can specify a list of IPCam cameras as a
            name->dict mapping. The keys will be unique names used to identify your
            cameras, the values will contain dictionaries containing `host, `port`,
            `username`, `password`, `timeout` and `ssl` attributes for each camera.
        """
        super().__init__(**kwargs)
        self.cameras: List[AndroidIpcam] = []
        self._camera_name_to_idx: Dict[str, int] = {}

        if not cameras:
            camera = AndroidIpcam(name=host, host=host, port=port, username=username,
                                  password=password, timeout=timeout, ssl=ssl)
            self.cameras.append(camera)
            self._camera_name_to_idx[host] = 0
        else:
            for name, camera in cameras.items():
                camera = AndroidIpcam(name=name, host=camera['host'], port=camera.get('port', port),
                                      username=camera.get('username'), password=camera.get('password'),
                                      timeout=camera.get('timeout', timeout), ssl=camera.get('ssl', ssl))
                self._camera_name_to_idx[name] = len(self.cameras)
                self.cameras.append(camera)

    def _get_camera(self, camera: Union[int, str] = None) -> AndroidIpcam:
        if not camera:
            camera = 0

        if isinstance(camera, int):
            return self.cameras[camera]

        return self.cameras[self._camera_name_to_idx[camera]]

    def _exec(self, url: str, camera: Union[int, str] = None, *args, **kwargs) -> Union[Dict[str, Any], bool]:
        cam = self._get_camera(camera)
        url = cam.base_url + url
        response = requests.get(url, auth=cam.auth, timeout=cam.timeout, verify=False, *args, **kwargs)
        response.raise_for_status()

        if response.headers.get('content-type') == 'application/json':
            return response.json()

        return response.text.find('Ok') != -1

    @action
    def change_setting(self, key: str, value: Union[str, int, bool], camera: Union[int, str] = None) -> bool:
        """
        Change a setting.
        :param key: Setting name
        :param value: Setting value
        :param camera: Camera index or configured name
        :return: True on success, False otherwise
        """
        if isinstance(value, bool):
            payload = "on" if value else "off"
        else:
            payload = value

        return self._exec("settings/{key}?set={payload}".format(key=key, payload=payload), camera=camera)

    @action
    def status(self, camera: Union[int, str] = None) -> AndroidCameraStatusListResponse:
        """
        :param camera: Camera index or name (default: status of all the cameras)
        :return: True if the camera is available, False otherwise
        """
        cameras = self._camera_name_to_idx.keys() if camera is None else [camera]
        statuses = []

        for c in cameras:
            try:
                if isinstance(camera, int):
                    cam = self.cameras[c]
                else:
                    cam = self.cameras[self._camera_name_to_idx[c]]

                status_data = self._exec('status.json', params={'show_avail': 1}, camera=cam.name).get('curvals', {})
                status = AndroidCameraStatusResponse(
                    name=cam.name,
                    stream_url=cam.stream_url,
                    image_url=cam.image_url,
                    audio_url=cam.audio_url,
                    **{k: v for k, v in status_data.items()
                       if k in AndroidCameraStatusResponse.attrs})

                statuses.append(status)
            except Exception as e:
                self.logger.warning('Could not get the status of {}: {}'.format(c, str(e)))

        return AndroidCameraStatusListResponse(statuses)

    @action
    def set_front_facing_camera(self, activate: bool = True, camera: Union[int, str] = None) -> bool:
        """Enable/disable the front-facing camera."""
        return self.change_setting('ffc', activate, camera=camera)

    @action
    def set_torch(self, activate: bool = True, camera: Union[int, str] = None) -> bool:
        """Enable/disable the torch."""
        url = 'enabletorch' if activate else 'disabletorch'
        return self._exec(url, camera=camera)

    @action
    def set_focus(self, activate: bool = True, camera: Union[int, str] = None) -> bool:
        """Enable/disable the focus."""
        url = 'focus' if activate else 'nofocus'
        return self._exec(url, camera=camera)

    @action
    def start_recording(self, tag: Optional[str] = None, camera: Union[int, str] = None) -> bool:
        """Start recording."""
        params = {'force': 1}
        if tag:
            params['tag'] = tag

        return self._exec('startvideo', params=params, camera=camera)

    @action
    def stop_recording(self, camera: Union[int, str] = None) -> bool:
        """Stop recording."""
        return self._exec('stopvideo', params={'force': 1}, camera=camera)

    @action
    def take_picture(self, image_file: str, camera: Union[int, str] = None) -> AndroidCameraPictureResponse:
        """Take a picture and save it on the local device."""
        cam = self._get_camera(camera)
        image_file = os.path.abspath(os.path.expanduser(image_file))
        os.makedirs(os.path.dirname(image_file), exist_ok=True)
        response = requests.get(cam.image_url, auth=cam.auth, verify=False)
        response.raise_for_status()

        with open(image_file, 'wb') as f:
            f.write(response.content)
        return AndroidCameraPictureResponse(image_file=image_file)

    @action
    def set_night_vision(self, activate: bool = True, camera: Union[int, str] = None) -> bool:
        """Enable/disable night vision."""
        return self.change_setting('night_vision', activate, camera=camera)

    @action
    def set_overlay(self, activate: bool = True, camera: Union[int, str] = None) -> bool:
        """Enable/disable video overlay."""
        return self.change_setting('overlay', activate, camera=camera)

    @action
    def set_gps(self, activate: bool = True, camera: Union[int, str] = None) -> bool:
        """Enable/disable GPS."""
        return self.change_setting('gps_active', activate, camera=camera)

    @action
    def set_quality(self, quality: int = 100, camera: Union[int, str] = None) -> bool:
        """Set video quality."""
        return self.change_setting('quality', int(quality), camera=camera)

    @action
    def set_motion_detect(self, activate: bool = True, camera: Union[int, str] = None) -> bool:
        """Enable/disable motion detect."""
        return self.change_setting('motion_detect', activate, camera=camera)

    @action
    def set_orientation(self, orientation: str = 'landscape', camera: Union[int, str] = None) -> bool:
        """Set video orientation."""
        return self.change_setting('orientation', orientation, camera=camera)

    @action
    def set_zoom(self, zoom: float, camera: Union[int, str] = None) -> bool:
        """Set the zoom level."""
        return self._exec('settings/ptz', params={'zoom': float(zoom)}, camera=camera)

    @action
    def set_scenemode(self, scenemode: str = 'auto', camera: Union[int, str] = None) -> bool:
        """Set video orientation."""
        return self.change_setting('scenemode', scenemode, camera=camera)


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import re
import time

from platypush.backend import Backend
from platypush.message import Message
from platypush.message.event.wiimote import WiimoteEvent, \
    WiimoteConnectionEvent, WiimoteDisconnectionEvent

from platypush.message.request import Request


class WiimoteBackend(Backend):
    """
    Backend to communicate with a Nintendo WiiMote controller

    Triggers:

        * :class:`platypush.message.event.Wiimote.WiimoteEvent` \
            when the state of the Wiimote (battery, buttons, acceleration etc.) changes

    Requires:

        * **python3-wiimote** (follow instructions at https://github.com/azzra/python3-wiimote)
    """

    _wiimote = None
    _inactivity_timeout = 300
    _connection_attempts = 0
    _last_btn_event_time = 0
    _bdaddr = None


    def __init__(self, bdaddr=_bdaddr, inactivity_timeout=_inactivity_timeout,
                 *args, **kwargs):
        """
        :param bdaddr: If set, connect to this specific Wiimote physical address (example: 00:11:22:33:44:55)
        :type bdaddr: str

        :param inactivity_timeout: Number of seconds elapsed from the last Wiimote action before disconnecting the device (default: 300 seconds)
        :type inactivity_timeout: float
        """

        super().__init__(*args, **kwargs)
        self._bdaddr = bdaddr
        self._inactivity_timeout = inactivity_timeout


    def get_wiimote(self):
        import cwiid

        if not self._wiimote:
            if self._bdaddr:
                self._wiimote = cwiid.Wiimote(bdaddr=self._bdaddr)
            else:
                self._wiimote = cwiid.Wiimote()

            self._wiimote.enable(cwiid.FLAG_MOTIONPLUS)
            self._wiimote.rpt_mode = cwiid.RPT_ACC | cwiid.RPT_BTN | cwiid.RPT_MOTIONPLUS

            self.logger.info('WiiMote connected')
            self._last_btn_event_time = time.time()
            self.bus.post(WiimoteConnectionEvent())

        return self._wiimote

    def get_state(self):
        import cwiid
        wm = self.get_wiimote()
        state = wm.state
        parsed_state = {}

        # Get buttons
        all_btns = [attr for attr in dir(cwiid) if attr.startswith('BTN_')]
        parsed_state['buttons'] = { btn: True for btn in all_btns
                if state.get('buttons', 0) & getattr(cwiid, btn) != 0 }

        # Get LEDs
        all_leds = [attr for attr in dir(cwiid) if re.match('LED\d_ON', attr)]
        parsed_state['led'] = { led[:4]: True for led in all_leds
                if state.get('leds', 0) & getattr(cwiid, led) != 0 }

        # Get errors
        all_errs = [attr for attr in dir(cwiid) if attr.startswith('ERROR_')]
        parsed_state['error'] = { err: True for err in all_errs
                if state.get('errs', 0) & getattr(cwiid, err) != 0 }

        parsed_state['battery'] = round(state.get('battery', 0)/cwiid.BATTERY_MAX, 3)
        parsed_state['rumble'] = bool(state.get('rumble', 0))

        if 'acc' in state:
            parsed_state['acc'] = tuple(int(acc/5)*5 for acc in state['acc'])

        if 'motionplus' in state:
            parsed_state['motionplus'] = {
                'angle_rate': tuple(int(angle/100) for angle
                                    in state['motionplus']['angle_rate']),
                'low_speed': state['motionplus']['low_speed'],
            }

        return parsed_state


    def close(self):
        if not self._wiimote:
            return

        try:
            self._wiimote.close()
        except:
            pass

        self._wiimote = None
        self.bus.post(WiimoteDisconnectionEvent())


    def run(self):
        super().run()

        self._connection_attempts = 0
        last_state = {}
        self.logger.info('Initialized Wiimote backend')

        while not self.should_stop():
            try:
                state = self.get_state()
                changed_state = { k: state[k] for k in state.keys()
                                 if state[k] != last_state.get(k) }

                if changed_state:
                    self.bus.post(WiimoteEvent(**changed_state))

                if 'buttons' in changed_state:
                    self._last_btn_event_time = time.time()
                elif last_state and time.time() - \
                    self._last_btn_event_time >= self._inactivity_timeout:
                    self.logger.info('Wiimote disconnected upon timeout')
                    self.close()

                last_state = state
                time.sleep(0.1)
            except RuntimeError as e:
                if type(e) == RuntimeError and str(e) == 'Error opening wiimote connection':
                    if self._connection_attempts == 0:
                        self.logger.info('Press 1+2 to pair your WiiMote controller')
                else:
                    self.logger.exception(e)

                self.close()
                self._connection_attempts += 1

# vim:sw=4:ts=4:et:




import datetime
import time

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.google.fit import GoogleFitEvent
from platypush.utils import camel_case_to_snake_case


class GoogleFitBackend(Backend):
    """
    This backend will listen for new Google Fit events (e.g. new weight/height
    measurements, new fitness activities etc.) on the specified data streams and
    fire an event upon new data.

    Triggers:

        * :class:`platypush.message.event.google.fit.GoogleFitEvent` when a new
            data point is received on one of the registered streams.

    Requires:

        * The **google.fit** plugin
            (:class:`platypush.plugins.google.fit.GoogleFitPlugin`) enabled.
        * The **db** plugin (:class:`platypush.plugins.db`) configured
    """

    _default_poll_seconds = 60
    _default_user_id = 'me'
    _last_timestamp_varname = '_GOOGLE_FIT_LAST_TIMESTAMP_'

    def __init__(self, data_sources, user_id=_default_user_id,
                 poll_seconds=_default_poll_seconds, *args, **kwargs):
        """
        :param data_sources: Google Fit data source IDs to monitor. You can
            get a list of the available data sources through the
            :meth:`platypush.plugins.google.fit.get_data_sources` action
        :type data_sources: list[str]

        :param user_id: Google user ID to track (default: 'me')
        :type user_id: str

        :param poll_seconds: How often the backend will query the data sources
            for new data points (default: 60 seconds)
        :type poll_seconds: float
        """

        super().__init__(*args, **kwargs)

        self.data_sources = data_sources
        self.user_id = user_id
        self.poll_seconds = poll_seconds

    def run(self):
        super().run()
        self.logger.info('Started Google Fit backend on data sources {}'.format(
            self.data_sources))

        while not self.should_stop():
            try:
                for data_source in self.data_sources:
                    varname = self._last_timestamp_varname + data_source
                    last_timestamp = float(get_plugin('variable').
                                           get(varname).output.get(varname) or 0)

                    new_last_timestamp = last_timestamp
                    self.logger.info('Processing new entries from data source {}, last timestamp: {}'.
                                     format(data_source,
                                            str(datetime.datetime.fromtimestamp(last_timestamp))))

                    data_points = get_plugin('google.fit').get_data(
                        user_id=self.user_id, data_source_id=data_source).output
                    new_data_points = 0

                    for dp in data_points:
                        dp_time = dp.pop('startTime', 0)
                        if 'dataSourceId' in dp:
                            del dp['dataSourceId']

                        if dp_time > last_timestamp:
                            self.bus.post(GoogleFitEvent(
                                user_id=self.user_id, data_source_id=data_source,
                                data_type=dp.pop('dataTypeName'),
                                start_time=dp_time,
                                end_time=dp.pop('endTime'),
                                modified_time=dp.pop('modifiedTime'),
                                values=dp.pop('values'),
                                **{camel_case_to_snake_case(k): v
                                    for k, v in dp.items()}
                            ))

                            new_data_points += 1

                        new_last_timestamp = max(dp_time, new_last_timestamp)

                    last_timestamp = new_last_timestamp
                    self.logger.info('Got {} new entries from data source {}, last timestamp: {}'.
                                     format(new_data_points, data_source,
                                            str(datetime.datetime.fromtimestamp(last_timestamp))))

                    get_plugin('variable').set(**{varname: last_timestamp})
            except Exception as e:
                self.logger.warning('Exception while processing Fit data')
                self.logger.exception(e)
                continue

            time.sleep(self.poll_seconds)


# vim:sw=4:ts=4:et:



from platypush.backend.sensor import SensorBackend


class SensorArduinoBackend(SensorBackend):
    """
    This backend listens for new events from an Arduino with a Firmata-compatible firmware.

    Requires:

        * The :class:`platypush.plugins.arduino.ArduinoPlugin` plugin configured.

    """

    def __init__(self, **kwargs):
        super().__init__(plugin='arduino', **kwargs)


# vim:sw=4:ts=4:et:



import argparse
import httplib2
import os
import sys

from oauth2client import client
from oauth2client import tools
from oauth2client.file import Storage


def get_credentials_filename(*scopes):
    from platypush.config import Config

    scope_name = '-'.join([scope.split('/')[-1] for scope in scopes])
    credentials_dir = os.path.join(
        Config.get('workdir'), 'credentials', 'google')

    os.makedirs(credentials_dir, exist_ok=True)
    return os.path.join(credentials_dir, scope_name + '.json')


def get_credentials(scope):
    credentials_file = get_credentials_filename(*sorted(scope.split(' ')))
    if not os.path.exists(credentials_file):
        raise RuntimeError(('Credentials file {} not found. Generate it through:\n' +
                           '\tpython -m platypush.plugins.google.credentials "{}" ' +
                           '<path to client_secret.json>\n' +
                           '\t\t[--auth_host_name AUTH_HOST_NAME]\n' +
                           '\t\t[--noauth_local_webserver]\n' +
                           '\t\t[--auth_host_port [AUTH_HOST_PORT [AUTH_HOST_PORT ...]]]\n' +
                           '\t\t[--logging_level [DEBUG,INFO,WARNING,ERROR,CRITICAL]]\n').
                           format(credentials_file, scope))

    store = Storage(credentials_file)
    credentials = store.get()

    if not credentials or credentials.invalid:
        credentials.refresh(httplib2.Http())

    return credentials


def generate_credentials(client_secret_path, scope):
    credentials_file = get_credentials_filename(*sorted(scope.split(' ')))
    store = Storage(credentials_file)

    flow = client.flow_from_clientsecrets(client_secret_path, scope)
    flow.user_agent = 'Platypush'
    flow.access_type = 'offline'
    flags = argparse.ArgumentParser(parents=[tools.argparser]).parse_args()
    credentials = tools.run_flow(flow, store, flags)
    print('Storing credentials to ' + credentials_file)


def main():
    """
    Generates a Google API credentials file given client secret JSON and scopes.
    Usage::

        python -m platypush.plugins.google.credentials [client_secret.json location] [comma-separated list of scopes]
    """

    scope = sys.argv.pop(1) if len(sys.argv) > 1 \
              else input('Space separated list of OAuth scopes: ')

    client_secret_path = os.path.expanduser(
        sys.argv.pop(1) if len(sys.argv) > 1
        else input('Google credentials JSON file location: '))

    # Uncomment to force headless (no browser spawned) authentication
    # sys.argv.append('--noauth_local_webserver')

    generate_credentials(client_secret_path, scope)


if __name__ == '__main__':
    main()


# vim:sw=4:ts=4:et:




import base64
import os
import tempfile

from flask import Response, request, Blueprint, send_from_directory

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate, send_request

camera_ir_mlx90640 = Blueprint('camera.ir.mlx90640', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    camera_ir_mlx90640,
]


def get_feed(**args):
    try:
        while True:
            frame = send_request(action='camera.ir.mlx90640.capture', **args).output[0]
            frame = base64.decodebytes(frame.encode())
            yield (b'--frame\r\n'
                   b'Content-Type: image/jpeg\r\n\r\n' + frame + b'\r\n')
    finally:
        send_request(action='camera.ir.mlx90640.stop')


@camera_ir_mlx90640.route('/camera/ir/mlx90640/frame', methods=['GET'])
@authenticate()
def get_frame_route():
    f = tempfile.NamedTemporaryFile(prefix='ir_camera_frame_', suffix='.jpg', delete=False)
    args = {
        'grayscale': bool(int(request.args.get('grayscale', 0))),
        'scale_factor': int(request.args.get('scale_factor', 1)),
        'rotate': int(request.args.get('rotate', 0)),
        'output_file': f.name,
    }

    send_request(action='camera.ir.mlx90640.capture', **args)
    return send_from_directory(os.path.dirname(f.name),
                               os.path.basename(f.name))


@camera_ir_mlx90640.route('/camera/ir/mlx90640/stream', methods=['GET'])
@authenticate()
def get_feed_route():
    args = {
        'grayscale': bool(int(request.args.get('grayscale', 0))),
        'scale_factor': int(request.args.get('scale_factor', 1)),
        'rotate': int(request.args.get('rotate', 0)),
        'format': 'jpeg',
    }

    return Response(get_feed(**args),
                    mimetype='multipart/x-mixed-replace; boundary=frame')

# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class SoundEvent(Event):
    """ Base class for sound events """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SoundPlaybackStartedEvent(SoundEvent):
    """
    Event triggered when a new sound playback starts
    """

    def __init__(self, filename=None, *args, **kwargs):
        super().__init__(*args, filename=filename, **kwargs)


class SoundPlaybackStoppedEvent(SoundEvent):
    """
    Event triggered when the sound playback stops
    """

    def __init__(self, filename=None, *args, **kwargs):
        super().__init__(*args, filename=filename, **kwargs)


class SoundPlaybackPausedEvent(SoundEvent):
    """
    Event triggered when the sound playback pauses
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SoundRecordingStartedEvent(SoundEvent):
    """
    Event triggered when a new recording starts
    """

    def __init__(self, filename=None, *args, **kwargs):
        super().__init__(*args, filename=filename, **kwargs)


class SoundRecordingStoppedEvent(SoundEvent):
    """
    Event triggered when a sound recording stops
    """

    def __init__(self, filename=None, *args, **kwargs):
        super().__init__(*args, filename=filename, **kwargs)


class SoundRecordingPausedEvent(SoundEvent):
    """
    Event triggered when a sound recording pauses
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# vim:sw=4:ts=4:et:



import sys

from ..hook import EventHook

from platypush.config import Config
from platypush.context import get_backend
from platypush.message.event import Event


class EventProcessor(object):
    """ Event processor class. Checks an event against the configured
        rules and executes any matching event hooks """

    def __init__(self, hooks=None):
        """
        Params:
            hooks -- List of event hooks (default: any entry in the config
            named as event.hook.<hook_name> """

        if hooks is None:
            hooks = Config.get_event_hooks()

        self.hooks = []
        for (name, hook) in hooks.items():
            h = EventHook.build(name=name, hook=hook)
            self.hooks.append(h)

    @staticmethod
    def notify_web_clients(event):
        backends = Config.get_backends()
        if 'http' not in backends:
            return

        backend = get_backend('http')
        if backend:
            backend.notify_web_clients(event)

        backend = get_backend('websocket')
        if backend:
            backend.notify_web_clients(event)

    def process_event(self, event: Event):
        """ Processes an event and runs the matched hooks with the highest score """

        if not event.disable_web_clients_notification:
            self.notify_web_clients(event)

        matched_hooks = set()
        priority_hooks = set()
        max_score = -sys.maxsize
        max_priority = 0

        for hook in self.hooks:
            match = hook.matches_event(event)
            if match.is_match:
                if match.score > max_score:
                    matched_hooks = {hook}
                    max_score = match.score
                elif match.score == max_score:
                    matched_hooks.add(hook)

                if hook.priority > max_priority:
                    priority_hooks = {hook}
                elif hook.priority == max_priority and max_priority > 0:
                    priority_hooks.add(hook)

        matched_hooks.update(priority_hooks)
        for hook in matched_hooks:
            hook.run(event)


# vim:sw=4:ts=4:et:



from flask import Blueprint, request, redirect, make_response, abort

from platypush.backend.http.app import template_folder
from platypush.user import UserManager

logout = Blueprint('logout', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    logout,
]


@logout.route('/logout', methods=['GET', 'POST'])
def logout():
    """ Logout page """
    user_manager = UserManager()
    redirect_page = request.args.get('redirect', request.headers.get('Referer', '/login'))
    session_token = request.cookies.get('session_token')

    if not session_token:
        return abort(417, 'Not logged in')

    user, session = user_manager.authenticate_user_session(session_token)
    if not user:
        return abort(403, 'Invalid session token')

    redirect_target = redirect(redirect_page, 302)
    response = make_response(redirect_target)
    response.set_cookie('session_token', '', expires=0)
    return response


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.backend.bluetooth.scanner import BluetoothScannerBackend


class BluetoothBleScannerBackend(BluetoothScannerBackend):
    """
    This backend periodically scans for available bluetooth low-energy devices and returns events when a devices enter
    or exits the range.

    Triggers:

        * :class:`platypush.message.event.bluetooth.BluetoothDeviceFoundEvent` when a new bluetooth device is found.
        * :class:`platypush.message.event.bluetooth.BluetoothDeviceLostEvent` when a bluetooth device is lost.

    Requires:

        * The :class:`platypush.plugins.bluetooth.BluetoothBlePlugin` plugin working.

    """

    def __init__(self, interface: Optional[int] = None, scan_duration: int = 10, **kwargs):
        """
        :param interface: Bluetooth adapter name to use (default configured on the ``bluetooth.ble`` plugin if None).
        :param scan_duration:  How long the scan should run (default: 10 seconds).
        """
        super().__init__(plugin='bluetooth.ble', plugin_args={
            'interface': interface,
            'duration': scan_duration,
        }, **kwargs)


# vim:sw=4:ts=4:et:



import importlib
import inspect
import json
import pkgutil
import re
import threading

import platypush.backend
import platypush.plugins
import platypush.message.event
import platypush.message.response

from platypush.backend import Backend
from platypush.plugins import Plugin, action
from platypush.message.event import Event
from platypush.message.response import Response
from platypush.utils import get_decorators


# noinspection PyTypeChecker
class Model:
    def __str__(self):
        return json.dumps(dict(self), indent=2, sort_keys=True)

    def __repr__(self):
        return json.dumps(dict(self))

    @staticmethod
    def to_html(doc):
        try:
            import docutils.core
        except ImportError:
            # docutils not found
            return doc

        return docutils.core.publish_parts(doc, writer_name='html')['html_body']


class BackendModel(Model):
    def __init__(self, backend, prefix='', html_doc: bool = False):
        self.name = backend.__module__[len(prefix):]
        self.html_doc = html_doc
        self.doc = self.to_html(backend.__doc__) if html_doc and backend.__doc__ else backend.__doc__

    def __iter__(self):
        for attr in ['name', 'doc', 'html_doc']:
            yield attr, getattr(self, attr)


class PluginModel(Model):
    def __init__(self, plugin, prefix='', html_doc: bool = False):
        self.name = plugin.__module__[len(prefix):]
        self.html_doc = html_doc
        self.doc = self.to_html(plugin.__doc__) if html_doc and plugin.__doc__ else plugin.__doc__
        self.actions = {action_name: ActionModel(getattr(plugin, action_name), html_doc=html_doc)
                        for action_name in get_decorators(plugin, climb_class_hierarchy=True).get('action', [])}

    def __iter__(self):
        for attr in ['name', 'actions', 'doc', 'html_doc']:
            if attr == 'actions':
                # noinspection PyShadowingNames
                yield attr, {name: dict(action) for name, action in self.actions.items()},
            else:
                yield attr, getattr(self, attr)


class EventModel(Model):
    def __init__(self, event, html_doc: bool = False):
        self.package = event.__module__
        self.name = event.__name__
        self.html_doc = html_doc
        self.doc = self.to_html(event.__doc__) if html_doc and event.__doc__ else event.__doc__

    def __iter__(self):
        for attr in ['name', 'doc', 'html_doc']:
            yield attr, getattr(self, attr)


class ResponseModel(Model):
    def __init__(self, response, html_doc: bool = False):
        self.package = response.__module__
        self.name = response.__name__
        self.html_doc = html_doc
        self.doc = self.to_html(response.__doc__) if html_doc and response.__doc__ else response.__doc__

    def __iter__(self):
        for attr in ['name', 'doc', 'html_doc']:
            yield attr, getattr(self, attr)


class ActionModel(Model):
    # noinspection PyShadowingNames
    def __init__(self, action, html_doc: bool = False):
        self.name = action.__name__
        self.doc, argsdoc = self._parse_docstring(action.__doc__, html_doc=html_doc)
        self.args = {}
        self.has_kwargs = False

        for arg in list(inspect.signature(action).parameters.values())[1:]:
            if arg.kind == arg.VAR_KEYWORD:
                self.has_kwargs = True
                continue

            self.args[arg.name] = {
                'default': arg.default if not issubclass(arg.default.__class__, type) else None,
                'doc': argsdoc.get(arg.name)
            }

    @classmethod
    def _parse_docstring(cls, docstring: str, html_doc: bool = False):
        new_docstring = ''
        params = {}
        cur_param = None
        cur_param_docstring = ''

        if not docstring:
            return None, {}

        for line in docstring.split('\n'):
            m = re.match(r'^\s*:param ([^:]+):\s*(.*)', line)
            if m:
                if cur_param:
                    params[cur_param] = cls.to_html(cur_param_docstring) if html_doc else cur_param_docstring

                cur_param = m.group(1)
                cur_param_docstring = m.group(2)
            elif re.match(r'^\s*:[^:]+:\s*.*', line):
                continue
            else:
                if cur_param:
                    if not line.strip():
                        params[cur_param] = cls.to_html(cur_param_docstring) if html_doc else cur_param_docstring
                        cur_param = None
                        cur_param_docstring = ''
                    else:
                        cur_param_docstring += '\n' + line.strip()
                else:
                    new_docstring += line.rstrip() + '\n'

        if cur_param:
            params[cur_param] = cls.to_html(cur_param_docstring) if html_doc else cur_param_docstring

        return new_docstring.strip() if not html_doc else cls.to_html(new_docstring), params

    def __iter__(self):
        for attr in ['name', 'args', 'doc', 'has_kwargs']:
            yield attr, getattr(self, attr)


class InspectPlugin(Plugin):
    """
    This plugin can be used to inspect platypush plugins and backends

    Requires:

        * **docutils** (``pip install docutils``) - optional, for HTML doc generation

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._plugins = {}
        self._backends = {}
        self._events = {}
        self._responses = {}
        self._plugins_lock = threading.RLock()
        self._backends_lock = threading.RLock()
        self._events_lock = threading.RLock()
        self._responses_lock = threading.RLock()
        self._html_doc = False

    def _init_plugins(self):
        package = platypush.plugins
        prefix = package.__name__ + '.'

        for _, modname, _ in pkgutil.walk_packages(path=package.__path__,
                                                   prefix=prefix,
                                                   onerror=lambda x: None):
            # noinspection PyBroadException
            try:
                module = importlib.import_module(modname)
            except:
                continue

            for _, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and issubclass(obj, Plugin):
                    model = PluginModel(plugin=obj, prefix=prefix, html_doc=self._html_doc)
                    if model.name:
                        self._plugins[model.name] = model

    def _init_backends(self):
        package = platypush.backend
        prefix = package.__name__ + '.'

        for _, modname, _ in pkgutil.walk_packages(path=package.__path__,
                                                   prefix=prefix,
                                                   onerror=lambda x: None):
            # noinspection PyBroadException
            try:
                module = importlib.import_module(modname)
            except:
                continue

            for _, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and issubclass(obj, Backend):
                    model = BackendModel(backend=obj, prefix=prefix, html_doc=self._html_doc)
                    if model.name:
                        self._backends[model.name] = model

    def _init_events(self):
        package = platypush.message.event
        prefix = package.__name__ + '.'

        for _, modname, _ in pkgutil.walk_packages(path=package.__path__,
                                                   prefix=prefix,
                                                   onerror=lambda x: None):
            # noinspection PyBroadException
            try:
                module = importlib.import_module(modname)
            except:
                continue

            for _, obj in inspect.getmembers(module):
                if type(obj) == Event:
                    continue

                if inspect.isclass(obj) and issubclass(obj, Event):
                    event = EventModel(event=obj, html_doc=self._html_doc)
                    if event.package not in self._events:
                        self._events[event.package] = {event.name: event}
                    else:
                        self._events[event.package][event.name] = event

    def _init_responses(self):
        package = platypush.message.response
        prefix = package.__name__ + '.'

        for _, modname, _ in pkgutil.walk_packages(path=package.__path__,
                                                   prefix=prefix,
                                                   onerror=lambda x: None):
            # noinspection PyBroadException
            try:
                module = importlib.import_module(modname)
            except:
                continue

            for _, obj in inspect.getmembers(module):
                if type(obj) == Response:
                    continue

                if inspect.isclass(obj) and issubclass(obj, Response):
                    response = ResponseModel(response=obj, html_doc=self._html_doc)
                    if response.package not in self._responses:
                        self._responses[response.package] = {response.name: response}
                    else:
                        self._responses[response.package][response.name] = response

    @action
    def get_all_plugins(self, html_doc: bool = None):
        """
        :param html_doc: If True then the docstring will be parsed into HTML (default: False)
        """
        with self._plugins_lock:
            if not self._plugins or (html_doc is not None and html_doc != self._html_doc):
                self._html_doc = html_doc
                self._init_plugins()

            return json.dumps({
                name: dict(plugin)
                for name, plugin in self._plugins.items()
            })

    @action
    def get_all_backends(self, html_doc: bool = None):
        """
        :param html_doc: If True then the docstring will be parsed into HTML (default: False)
        """
        with self._backends_lock:
            if not self._backends or (html_doc is not None and html_doc != self._html_doc):
                self._html_doc = html_doc
                self._init_backends()

            return json.dumps({
                name: dict(backend)
                for name, backend in self._backends.items()
            })

    @action
    def get_all_events(self, html_doc: bool = None):
        """
        :param html_doc: If True then the docstring will be parsed into HTML (default: False)
        """
        with self._events_lock:
            if not self._events or (html_doc is not None and html_doc != self._html_doc):
                self._html_doc = html_doc
                self._init_events()

            return json.dumps({
                package: {
                    name: dict(event)
                    for name, event in self._events[package].items()
                }
                for package, events in self._events.items()
            })

    @action
    def get_all_responses(self, html_doc: bool = None):
        """
        :param html_doc: If True then the docstring will be parsed into HTML (default: False)
        """
        with self._responses_lock:
            if not self._responses or (html_doc is not None and html_doc != self._html_doc):
                self._html_doc = html_doc
                self._init_responses()

            return json.dumps({
                package: {
                    name: dict(event)
                    for name, event in self._responses[package].items()
                }
                for package, events in self._responses.items()
            })


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import os

from platypush.plugins import Plugin


class GooglePlugin(Plugin):
    """
    Executes calls to the Google APIs using the google-api-python-client.
    This class is extended by ``GoogleMailPlugin``, ``GoogleCalendarPlugin`` etc.
    In order to use Google services (like GMail, Maps, Calendar etc.) with
    your account you need to:

        1. Create your Google application, if you don't have one already, on
        the developers console, https://console.developers.google.com

        2. Click on "Credentials", then "Create credentials" -> "OAuth client ID"

        3 Select "Other", enter whichever description you like, and create

        4. Click on the "Download JSON" icon next to your newly created client ID

        5. Generate a credentials file for the needed scope::

            python -m platypush.plugins.google.credentials 'https://www.googleapis.com/auth/gmail.compose' ~/client_secret.json

    Requires:

        * **google-api-python-client** (``pip install google-api-python-client``)
        * **oauth2client** (``pip install oauth2client``)

    """

    def __init__(self, scopes=None, *args, **kwargs):
        """
        Initialized the Google plugin with the required scopes.

        :param scopes: List of required scopes
        :type scopes: list
        """

        from platypush.plugins.google.credentials import get_credentials
        super().__init__(*args, **kwargs)
        self._scopes = scopes or []

        if self._scopes:
            scopes = ' '.join(sorted(self._scopes))
            self.credentials = {
                scopes: get_credentials(scopes)
            }
        else:
            self.credentials = {}


    def get_service(self, service, version, scopes=None):
        import httplib2
        from apiclient import discovery

        if scopes is None:
            scopes = getattr(self, 'scopes') if hasattr(self, 'scopes') else []

        scopes = ' '.join(sorted(scopes))
        credentials = self.credentials[scopes]
        http = credentials.authorize(httplib2.Http())
        return discovery.build(service, version, http=http, cache_discovery=False)


# vim:sw=4:ts=4:et:



import base64
import json
import socket

from typing import Optional, Union

from platypush.plugins import Plugin, action


class TcpPlugin(Plugin):
    """
    Plugin for raw TCP communications.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._sockets = {}

    def _connect(self, host: str, port: int, timeout: Optional[float] = None) -> socket.socket:
        sd = self._sockets.get((host, port))
        if sd:
            return sd

        sd = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if timeout:
            sd.settimeout(timeout)
        sd.connect((host, port))
        self._sockets[(host, port)] = sd
        return sd

    @action
    def connect(self, host: str, port: int, timeout: Optional[float] = None):
        """
        Open a TCP connection.

        :param host: Host IP/name.
        :param port: TCP port.
        :param timeout: Connection timeout in seconds (default: None).
        """
        self._connect(host, port, timeout)

    @action
    def close(self, host: str, port: int):
        """
        Close an active TCP connection.

        :param host: Host IP/name.
        :param port: TCP port.
        """
        sd = self._sockets.get((host, port))
        if not sd:
            self.logger.warning('Not connected to ({}, {})'.format(host, port))
            return

        sd.close()

    @action
    def send(self, data: Union[bytes, str], host: str, port: int, binary: bool = False,
             timeout: Optional[float] = None, recv_response: bool = False, **recv_opts):
        """
        Send data over a TCP connection. If the connection isn't active it will be created.

        :param data: Data to be sent, as bytes or string.
        :param host: Host IP/name.
        :param port: TCP port.
        :param binary: If set to True and ``data`` is a string then will be treated as base64-encoded binary input.
        :param timeout: Connection timeout in seconds (default: None).
        :param recv_response: If True then the action will wait for a response from the server before closing the
            connection. Note that ``recv_opts`` must be specified in this case - at least ``length``.
        """
        if isinstance(data, list) or isinstance(data, dict):
            data = json.dumps(data)
        if isinstance(data, str):
            data = data.encode()
            if binary:
                data = base64.decodebytes(data)

        sd = self._connect(host, port, timeout)

        try:
            sd.send(data)
            if recv_response:
                recv_opts.update({
                    'host': host,
                    'port': port,
                    'timeout': timeout,
                    'binary': binary,
                })

                return self.recv(**recv_opts)
        finally:
            self.close(host, port)

    @action
    def recv(self, length: int, host: str, port: int, binary: bool = False, timeout: Optional[float] = None) -> str:
        """
        Receive data from a TCP connection. If the connection isn't active it will be created.

        :param length: Maximum number of bytes to be received.
        :param host: Host IP/name.
        :param port: TCP port.
        :param binary: If set to True then the output will be base64-encoded, otherwise decoded as string.
        :param timeout: Connection timeout in seconds (default: None).
        """
        sd = self._connect(host, port, timeout)

        try:
            data = sd.recv(length)
            if binary:
                data = base64.encodebytes(data).decode()
            else:
                data = data.decode()

            return data
        finally:
            self.close(host, port)


# vim:sw=4:ts=4:et:



from typing import Optional, Union

from platypush.message.event import Event


class ZeroborgEvent(Event):
    pass


class ZeroborgDriveEvent(ZeroborgEvent):
    def __init__(self, motors: Union[list, tuple], direction: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, motors=motors, direction=direction, **kwargs)


class ZeroborgStopEvent(ZeroborgEvent):
    pass


# vim:sw=4:ts=4:et:



import datetime
import requests
from typing import List, Dict, Any, Optional, Union, Tuple

from platypush.plugins import Plugin, action


class FoursquarePlugin(Plugin):
    """
    Plugin to interact with the `Foursquare Places API <https://developer.foursquare.com/docs/api>`_.

    In order to enable the Foursquare API on your account you need to:

        - Create a new app on the `Foursquare developers website <https://foursquare.com/developers/signup>`_.
        - Copy the ``client_id`` and ``client_secret``.
        - Add a redirect URL. It must point to a valid IP/hostname with a web server running, even if it runs
            locally. You can also use the local URL of the platypush web server - e.g. http://192.168.1.2:8008/.
        - Open the following URL: ``https://foursquare.com/oauth2/authenticate?client_id=CLIENT_ID&response_type=token&redirect_uri=REDIRECT_URI``.
            Replace ``CLIENT_ID`` and ``REDIRECT_URI`` with the parameters from your app.
        - Allow the application. You will be redirected to the URL you provided. Copy the ``access_token`` provided in
            the URL.

    """

    api_base_url = 'https://api.foursquare.com/v2'

    def __init__(self, access_token: str, **kwargs):
        """
        :param access_token:
        """
        super().__init__(**kwargs)
        self.access_token = access_token

    def _get_url(self, endpoint):
        return '{url}/{endpoint}?oauth_token={token}&v={version}'.format(
            url=self.api_base_url, endpoint=endpoint, token=self.access_token,
            version=datetime.date.today().strftime('%Y%m%d'),
        )

    @action
    def get_checkins(self) -> List[Dict[str, Any]]:
        """
        Get the list of check-ins of the current user.
        :return: A list of checkins, as returned by the Foursquare API.
        """
        url = self._get_url('users/self/checkins')
        return requests.get(url).json().get('response', {}).get('checkins', {}).get('items', [])

    # noinspection DuplicatedCode
    @action
    def search(self,
               latitude: Optional[float] = None,
               longitude: Optional[float] = None,
               altitude: Optional[float] = None,
               latlng_accuracy: Optional[float] = None,
               altitude_accuracy: Optional[float] = None,
               near: Optional[str] = None,
               query: Optional[str] = None,
               limit: Optional[int] = None,
               url: Optional[int] = None,
               categories: Optional[List[str]] = None,
               radius: Optional[int] = None,
               sw: Optional[Union[Tuple[float], List[float]]] = None,
               ne: Optional[Union[Tuple[float], List[float]]] = None,) -> List[Dict[str, Any]]:
        """
        Search for venues.

        :param latitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param longitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param near: Search near this place (e.g. "Chicago, IL" or "Amsterdam, NL"). Note either
            ``latitude, longitude`` or ``near`` should be provided.
        :param altitude: Search near this altitude in meters.
        :param latlng_accuracy:  Latitude/longitude accuracy in meters.
        :param altitude_accuracy: Altitude accuracy in meters.
        :param query: Search query (e.g. "coffee shops" or "restaurants").
        :param limit: Maximum number of results.
        :param url: Venue URL to search.
        :param categories: List of ``category IDs <https://developer.foursquare.com/docs/resources/categories>``_
            to be searched.
        :param radius: Search radius in meters.
        :param sw: South/west boundary box as a ``[latitude, longitude]`` pair.
        :param ne: North/east boundary box as a ``[latitude, longitude]`` pair.
        :return: A list of venues, as returned by the Foursquare API.
        """
        assert (latitude and longitude) or near, 'Specify either latitude/longitude or near'
        args = {}

        if latitude and longitude:
            args['ll'] = ','.join([str(latitude), str(longitude)])
        if near:
            args['near'] = near
        if altitude:
            args['alt'] = altitude
        if latlng_accuracy:
            args['llAcc'] = latlng_accuracy
        if altitude_accuracy:
            args['altAcc'] = altitude_accuracy
        if query:
            args['query'] = query
        if limit:
            args['limit'] = int(limit)
        if url:
            args['url'] = url
        if categories:
            args['categoryId'] = ','.join(categories)
        if radius:
            args['radius'] = radius
        if sw:
            args['sw'] = sw
        if ne:
            args['ne'] = ne

        url = self._get_url('venues/search')
        return requests.get(url, params=args).json().get('response', {}).get('venues', [])

    # noinspection DuplicatedCode
    @action
    def explore(self,
                latitude: Optional[float] = None,
                longitude: Optional[float] = None,
                altitude: Optional[float] = None,
                latlng_accuracy: Optional[float] = None,
                altitude_accuracy: Optional[float] = None,
                section: Optional[str] = None,
                near: Optional[str] = None,
                query: Optional[str] = None,
                limit: Optional[int] = None,
                categories: Optional[List[str]] = None,
                radius: Optional[int] = None,
                open_now: bool = True,
                sort_by_distance: Optional[bool] = None,
                sort_by_popularity: Optional[bool] = None,
                price: Optional[List[int]] = None,
                saved: Optional[bool] = None) -> List[Dict[str, Any]]:
        """
        Explore venues around a location.

        :param latitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param longitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param near: Search near this place (e.g. "Chicago, IL" or "Amsterdam, NL"). Note either
            ``latitude, longitude`` or ``near`` should be provided.
        :param altitude: Search near this altitude in meters.
        :param latlng_accuracy:  Latitude/longitude accuracy in meters.
        :param altitude_accuracy: Altitude accuracy in meters.
        :param section: Section to search. Supported values:

            - food
            - drinks
            - coffee
            - shops
            - arts
            - outdoors
            - sights
            - trending
            - nextVenues

        :param query: Search query (e.g. "coffee shops" or "restaurants"). The parameter has no effect if
            ``section`` is specified.
        :param limit: Maximum number of results.
        :param categories: List of ``category IDs <https://developer.foursquare.com/docs/resources/categories>``_
            to be searched.
        :param radius: Search radius in meters.
        :param open_now: Filter by open/not open now.
        :param sort_by_distance: Sort by distance.
        :param sort_by_popularity: Sort by popularity
        :param price: Price ranges, within the range ``[1,2,3,4]``.
        :param saved: Filter by saved/unsaved venues.

        :return: A list of venues, as returned by the Foursquare API.
        """
        assert (latitude and longitude) or near, 'Specify either latitude/longitude or near'
        args = {}

        if latitude and longitude:
            args['ll'] = ','.join([str(latitude), str(longitude)])
        if near:
            args['near'] = near
        if altitude:
            args['alt'] = altitude
        if latlng_accuracy:
            args['llAcc'] = latlng_accuracy
        if altitude_accuracy:
            args['altAcc'] = altitude_accuracy
        if section:
            args['section'] = section
        if query:
            args['query'] = query
        if categories:
            args['categoryId'] = ','.join(categories)
        if limit:
            args['limit'] = int(limit)
        if radius:
            args['radius'] = radius
        if open_now is not None:
            args['openNow'] = int(open_now)
        if sort_by_distance is not None:
            args['sortByDistance'] = int(sort_by_distance)
        if sort_by_popularity is not None:
            args['sortByPopularity'] = sort_by_popularity
        if saved is not None:
            args['saved'] = int(saved)
        if price:
            args['price'] = ','.join([str(p) for p in price])

        url = self._get_url('venues/explore')
        return requests.get(url, params=args).json().get('response', {}).get('venues', [])

    @action
    def trending(self,
                 latitude: Optional[float] = None,
                 longitude: Optional[float] = None,
                 near: Optional[str] = None,
                 limit: Optional[int] = None,
                 radius: Optional[int] = None) -> List[Dict[str, Any]]:
        """
        Get the trending venues around a location.

        :param latitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param longitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param near: Search near this place (e.g. "Chicago, IL" or "Amsterdam, NL"). Note either
            ``latitude, longitude`` or ``near`` should be provided.
        :param limit: Maximum number of results.
        :param radius: Search radius in meters.

        :return: A list of venues, as returned by the Foursquare API.
        """
        assert (latitude and longitude) or near, 'Specify either latitude/longitude or near'
        args = {}

        if latitude and longitude:
            args['ll'] = ','.join([str(latitude), str(longitude)])
        if near:
            args['near'] = near
        if limit:
            args['limit'] = int(limit)
        if radius:
            args['radius'] = radius

        url = self._get_url('venues/trending')
        return requests.get(url, params=args).json().get('response', {}).get('venues', [])

    @staticmethod
    def _parse_time(t):
        if isinstance(t, int) or isinstance(t, float):
            return datetime.datetime.fromtimestamp(t)
        if isinstance(t, str):
            return datetime.datetime.fromisoformat(t)

        assert isinstance(t, datetime.datetime), 'Cannot parse object of type {} into datetime: {}'.format(
            type(t), t)
        return t

    @action
    def time_series(self,
                    venue_id: Union[str, List[str]],
                    start_at: Union[int, float, datetime.datetime, str],
                    end_at: Union[int, float, datetime.datetime, str]) -> List[Dict[str, Any]]:
        """
        Get the visitors stats about one or multiple venues over a time range. The user must be a manager of
        those venues.

        :param venue_id: Venue ID or list of IDs to get the stats for.
        :param start_at: Stats start time. Can be a UNIX timestamp, a datetime object or an ISO format datetime.
        :param end_at: Stats end time. Can be a UNIX timestamp, a datetime object or an ISO format datetime.

        :return: A list of venues, as returned by the Foursquare API.
        """
        if isinstance(venue_id, list):
            venue_id = ','.join(venue_id)

        args = {
            'venueId': venue_id,
            'startAt': self._parse_time(start_at),
            'endAt': self._parse_time(end_at),
        }

        url = self._get_url('venues/timeseries')
        return requests.get(url, params=args).json().get('response', {}).get('venues', [])

    @action
    def stats(self,
              venue_id: str,
              start_at: Union[int, float, datetime.datetime, str],
              end_at: Union[int, float, datetime.datetime, str]) -> List[Dict[str, Any]]:
        """
        Get the stats about a venue over a time range. The user must be a manager of that venue.

        :param venue_id: Venue ID.
        :param start_at: Stats start time. Can be a UNIX timestamp, a datetime object or an ISO format datetime.
        :param end_at: Stats end time. Can be a UNIX timestamp, a datetime object or an ISO format datetime.

        :return: A list of venues, as returned by the Foursquare API.
        """
        args = {
            'startAt': self._parse_time(start_at),
            'endAt': self._parse_time(end_at),
        }

        url = self._get_url('venues/{}/stats'.format(venue_id))
        return requests.get(url, params=args).json().get('response', {}).get('venues', [])

    @action
    def managed(self) -> List[Dict[str, Any]]:
        """
        Get the list of venues managed by the user.
        :return: A list of venues, as returned by the Foursquare API.
        """
        url = self._get_url('venues/managed')
        return requests.get(url).json().get('response', {}).get('venues', []).get('items', [])

    @action
    def checkin(self,
                venue_id: str,
                latitude: Optional[float] = None,
                longitude: Optional[float] = None,
                altitude: Optional[float] = None,
                latlng_accuracy: Optional[float] = None,
                altitude_accuracy: Optional[float] = None,
                shout: Optional[str] = None,
                broadcast: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Create a new check-in.

        :param venue_id: ID of the venue to check-in.
        :param latitude: Check-in latitude.
        :param longitude: Check-in longitude.
        :param altitude: Check-in altitude.
        :param latlng_accuracy:  Latitude/longitude accuracy in meters.
        :param altitude_accuracy: Altitude accuracy in meters.
        :param shout: Add a custom message to the check-in.
        :param broadcast: List of Visibility/share types of the check-in. Default: ``public``. Possible values are:

            - ``private``
            - ``public``
            - ``followers``
            - ``facebook``
            - ``twitter``

        :return: Foursquare API response.
        """
        args = {'venueId': venue_id}
        if latitude and longitude:
            args['ll'] = ','.join([str(latitude), str(longitude)])
        if altitude:
            args['alt'] = altitude
        if latlng_accuracy:
            args['llAcc'] = latlng_accuracy
        if altitude_accuracy:
            args['altAcc'] = altitude_accuracy
        if shout:
            args['shout'] = shout
        if broadcast:
            args['broadcast'] = ','.join(broadcast) if isinstance(broadcast, list) else broadcast

        url = self._get_url('checkins/add')
        return requests.post(url, data=args).json().get('response', {}).get('checkin', {})


# vim:sw=4:ts=4:et:



import json
import time
from typing import Optional

from platypush.backend import Backend
from platypush.message.event.pushbullet import PushbulletEvent


class PushbulletBackend(Backend):
    """
    This backend will listen for events on a Pushbullet (https://pushbullet.com)
    channel and propagate them to the bus. This backend is quite useful if you
    want to synchronize events and actions with your mobile phone (through the
    Pushbullet app and/or through Tasker), synchronize clipboards, send pictures
    and files to other devices etc. You can also wrap Platypush messages as JSON
    into a push body to execute them.

    Triggers:

        * :class:`platypush.message.event.pushbullet.PushbulletEvent` if a new push is received

    Requires:

        * **requests** (``pip install requests``)
        * **pushbullet.py** (``pip install git+https://github.com/rbrcsk/pushbullet.py``)
    """

    def __init__(self, token: str, device: str = 'Platypush', proxy_host: Optional[str] = None,
                 proxy_port: Optional[int] = None, **kwargs):
        """
        :param token: Your Pushbullet API token, see https://docs.pushbullet.com/#authentication
        :param device: Name of the virtual device for Platypush (default: Platypush)
        :param proxy_host: HTTP proxy host (default: None)
        :param proxy_port: HTTP proxy port (default: None)
        """
        super().__init__(**kwargs)

        self.token = token
        self.device_name = device
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.device = None
        self.pb_device_id = None
        self.pb = None
        self.listener = None

    def _initialize(self):
        # noinspection PyPackageRequirements
        from pushbullet import Pushbullet
        self.pb = Pushbullet(self.token)

        # noinspection PyBroadException
        try:
            self.device = self.pb.get_device(self.device_name)
        except:
            self.device = self.pb.new_device(self.device_name)

        self.pb_device_id = self.get_device_id()

    def _get_latest_push(self):
        t = int(time.time()) - 5
        pushes = self.pb.get_pushes(modified_after=str(t), limit=1)
        if pushes:
            return pushes[0]

    def on_push(self):
        def callback(data):
            try:
                # Parse the push
                try:
                    data = json.loads(data) if isinstance(data, str) else data
                except Exception as e:
                    self.logger.exception(e)
                    return

                # If it's a push, get it
                if data['type'] == 'tickle' and data['subtype'] == 'push':
                    push = self._get_latest_push()
                elif data['type'] == 'push':
                    push = data['push']
                else:
                    return  # Not a push notification

                if not push:
                    return

                # Post an event, useful to react on mobile notifications if
                # you enabled notification mirroring on your PushBullet app
                event = PushbulletEvent(**push)
                self.on_message(event)

                if 'body' not in push:
                    return
                self.logger.debug('Received push: {}'.format(push))

                body = push['body']
                try:
                    body = json.loads(body)
                    self.on_message(body)
                except Exception as e:
                    self.logger.debug(('Unexpected message received on the ' +
                                       'Pushbullet backend: {}. Message: {}')
                                      .format(str(e), body))

            except Exception as e:
                self.logger.exception(e)
                return

        return callback

    def get_device_id(self):
        # noinspection PyBroadException
        try:
            return self.pb.get_device(self.device_name).device_iden
        except Exception:
            device = self.pb.new_device(self.device_name, model='Platypush virtual device',
                                        manufacturer='platypush', icon='system')

            self.logger.info('Created Pushbullet device {}'.format(
                self.device_name))

            return device.device_iden

    def close(self):
        if self.listener:
            self.listener.close()

    def on_stop(self):
        return self.close()

    def run(self):
        # noinspection PyPackageRequirements
        from pushbullet import Listener
        super().run()
        initialized = False

        while not initialized:
            try:
                self._initialize()
                initialized = True
            except Exception as e:
                self.logger.exception(e)
                self.logger.error('Pushbullet initialization error: {}'.format(str(e)))
                time.sleep(30)

        self.logger.info('Initialized Pushbullet backend - device_id: {}'
                         .format(self.device_name))

        self.listener = Listener(account=self.pb, on_push=self.on_push(),
                                 http_proxy_host=self.proxy_host,
                                 http_proxy_port=self.proxy_port)

        self.listener.run_forever()

# vim:sw=4:ts=4:et:



import threading
from typing import Optional, Dict, Any, Tuple

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.plugins.tts import TtsPlugin


class AssistantBackend(Backend):
    def __init__(self, tts_plugin: Optional[str] = None, tts_args: Optional[Dict[str, Any]] = None, **kwargs):
        """
        Default assistant backend constructor.

        :param tts_plugin: If set, and if the assistant returns the processed response as text, then the processed
            response will be played through the selected text-to-speech plugin (can be e.g. "``tts``",
            "``tts.google``" or any other implementation of :class:`platypush.plugins.tts.TtsPlugin`).
        :param tts_args: Extra parameters to pass to the ``say`` method of the selected TTS plugin (e.g.
            language, voice or gender).
        """
        super().__init__(**kwargs)
        self._detection_paused = threading.Event()
        self.tts_plugin = tts_plugin
        self.tts_args = tts_args or {}

    def pause_detection(self):
        self._detection_paused.set()

    def resume_detection(self):
        self._detection_paused.clear()

    def is_detecting(self):
        return not self._detection_paused.is_set()

    def _get_tts_plugin(self) -> Tuple[Optional[TtsPlugin], Dict[str, Any]]:
        return get_plugin(self.tts_plugin) if self.tts_plugin else None, self.tts_args


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class TorrentEvent(Event):
    """
    Base class for torrent events
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TorrentQueuedEvent(TorrentEvent):
    """
    Event triggered upon when a new torrent transfer is queued
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


class TorrentDownloadedMetadataEvent(TorrentEvent):
    """
    Event triggered upon torrent metadata download completed
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


class TorrentDownloadStartEvent(TorrentEvent):
    """
    Event triggered upon torrent download start
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


class TorrentSeedingStartEvent(TorrentEvent):
    """
    Event triggered upon torrent seeding start
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


class TorrentDownloadProgressEvent(TorrentEvent):
    """
    Event triggered upon torrent download progress
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


class TorrentStateChangeEvent(TorrentEvent):
    """
    Event triggered upon torrent state change
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


class TorrentPausedEvent(TorrentEvent):
    """
    Event triggered when a torrent transfer is paused
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


class TorrentResumedEvent(TorrentEvent):
    """
    Event triggered when a torrent transfer is resumed
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


class TorrentDownloadCompletedEvent(TorrentEvent):
    """
    Event triggered upon torrent state change
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


class TorrentDownloadStopEvent(TorrentEvent):
    """
    Event triggered when a torrent transfer is stopped
    """
    def __init__(self, url, *args, **kwargs):
        super().__init__(*args, url=url, **kwargs)


# vim:sw=4:ts=4:et:



from platypush.message.event.http import HttpEvent


class NewFeedEvent(HttpEvent):
    """
    Event triggered when a monitored RSS feed has some new content
    """

    def __init__(self, request, response: list, source_id=None, title=None,
                 digest_format=None, digest_filename=None, *args, **kwargs):
        """
        :param request: Original request
        :param response: Received response
        :param source_id: ID of the source that generated the event
        :param title: Title of the new element
        :param digest_format: Format of the digest - either 'html' or 'pdf', if set
        :param digest_filename: File name of the digest, if it was dumped to file
        """

        super().__init__(request=request, response=response, source_id=source_id,
                         digest_format=digest_format, title=title,
                         digest_filename=digest_filename, *args, **kwargs)


# vim:sw=4:ts=4:et:



import os
import urllib.parse

from platypush.context import get_bus, get_plugin
from platypush.plugins.media import PlayerState, MediaPlugin
from platypush.message.event.media import MediaPlayEvent, MediaPlayRequestEvent, \
    MediaPauseEvent, MediaStopEvent, MediaSeekEvent, MediaVolumeChangedEvent, \
    MediaMuteChangedEvent, NewPlayingMediaEvent

from platypush.plugins import action


class MediaVlcPlugin(MediaPlugin):
    """
    Plugin to control vlc instances

    Requires:

        * **python-vlc** (``pip install python-vlc``)
        * **vlc** executable on your system
    """

    def __init__(self, args=None, fullscreen=False, volume=100, *argv, **kwargs):
        """
        Create the vlc wrapper.

        :param args: List of extra arguments to pass to the VLC executable (e.g.
            ``['--sub-language=en', '--snapshot-path=/mnt/snapshots']``)
        :type args: list[str]

        :param fullscreen: Set to True if you want media files to be opened in
            fullscreen by default (can be overridden by `.play()`) (default: False)
        :type fullscreen: bool

        :param volume: Default media volume (default: 100)
        :type volume: int
        """

        super().__init__(*argv, **kwargs)

        self._args = args or []
        self._instance = None
        self._player = None
        self._latest_seek = None
        self._default_fullscreen = fullscreen
        self._default_volume = volume
        self._on_stop_callbacks = []
        self._title = None
        self._filename = None

    @classmethod
    def _watched_event_types(cls):
        import vlc
        return [getattr(vlc.EventType, evt) for evt in [
            'MediaPlayerLengthChanged', 'MediaPlayerMediaChanged',
            'MediaDurationChanged', 'MediaPlayerMuted',
            'MediaPlayerUnmuted', 'MediaPlayerOpening', 'MediaPlayerPaused',
            'MediaPlayerPlaying', 'MediaPlayerPositionChanged',
            'MediaPlayerStopped', 'MediaPlayerTimeChanged', 'MediaStateChanged',
            'MediaPlayerForward', 'MediaPlayerBackward',
            'MediaPlayerEndReached', 'MediaPlayerTitleChanged',
            'MediaPlayerAudioVolume',
        ] if hasattr(vlc.EventType, evt)]

    def _init_vlc(self, resource):
        import vlc
        self._reset_state()

        for k, v in self._env.items():
            os.environ[k] = v

        self._instance = vlc.Instance(*self._args)
        self._player = self._instance.media_player_new(resource)

        for evt in self._watched_event_types():
            self._player.event_manager().event_attach(
                eventtype=evt, callback=self._event_callback())

    def _reset_state(self):
        self._latest_seek = None
        self._title = None
        self._filename = None

        if self._player:
            self._player.release()
            self._player = None

        if self._instance:
            self._instance.release()
            self._instance = None

    @staticmethod
    def _post_event(evt_type, **evt):
        bus = get_bus()
        bus.post(evt_type(player='local', plugin='media.vlc', **evt))

    def _event_callback(self):
        def callback(event):
            from vlc import EventType
            self.logger.debug('Received vlc event: {}'.format(event))

            if event.type == EventType.MediaPlayerPlaying:
                self._post_event(MediaPlayEvent, resource=self._get_current_resource())
            elif event.type == EventType.MediaPlayerPaused:
                self._post_event(MediaPauseEvent)
            elif event.type == EventType.MediaPlayerStopped or \
                    event.type == EventType.MediaPlayerEndReached:
                self._reset_state()
                self._post_event(MediaStopEvent)
                for cbk in self._on_stop_callbacks:
                    cbk()
            elif event.type == EventType.MediaPlayerTitleChanged:
                self._filename = event.u.filename
                self._title = event.u.new_title
                self._post_event(NewPlayingMediaEvent, resource=event.u.new_title)
            elif event.type == EventType.MediaPlayerMediaChanged:
                self._filename = event.u.filename
                self._title = event.u.new_title
                self._post_event(NewPlayingMediaEvent, resource=event.u.filename)
            elif event.type == EventType.MediaPlayerLengthChanged:
                self._post_event(NewPlayingMediaEvent, resource=self._get_current_resource())
            elif event.type == EventType.MediaPlayerTimeChanged:
                pos = float(event.u.new_time/1000)
                if self._latest_seek is None or \
                        abs(pos-self._latest_seek) > 5:
                    self._post_event(MediaSeekEvent, position=pos)
                self._latest_seek = pos
            elif event.type == EventType.MediaPlayerAudioVolume:
                self._post_event(MediaVolumeChangedEvent, volume=self._player.audio_get_volume())
            elif event.type == EventType.MediaPlayerMuted:
                self._post_event(MediaMuteChangedEvent, mute=True)
            elif event.type == EventType.MediaPlayerUnmuted:
                self._post_event(MediaMuteChangedEvent, mute=False)

        return callback

    @action
    def play(self, resource, subtitles=None, fullscreen=None, volume=None):
        """
        Play a resource.

        :param resource: Resource to play - can be a local file or a remote URL
        :type resource: str

        :param subtitles: Path to optional subtitle file
        :type subtitles: str

        :param fullscreen: Set to explicitly enable/disable fullscreen (default:
            `fullscreen` configured value or False)
        :type fullscreen: bool

        :param volume: Set to explicitly set the playback volume (default:
            `volume` configured value or 100)
        :type fullscreen: bool
        """

        get_bus().post(MediaPlayRequestEvent(resource=resource))
        resource = self._get_resource(resource)

        if resource.startswith('file://'):
            resource = resource[len('file://'):]

        self._init_vlc(resource)
        if subtitles:
            if subtitles.startswith('file://'):
                subtitles = subtitles[len('file://'):]
            self._player.video_set_subtitle_file(subtitles)

        self._player.play()
        if self.volume:
            self.set_volume(volume=self.volume)

        if fullscreen or self._default_fullscreen:
            self.set_fullscreen(True)

        if volume is not None or self._default_volume is not None:
            self.set_volume(volume if volume is not None
                            else self._default_volume)

        return self.status()

    @action
    def pause(self):
        """ Toggle the paused state """
        if not self._player:
            return None, 'No vlc instance is running'
        if not self._player.can_pause():
            return None, 'The specified media type cannot be paused'

        self._player.pause()
        return self.status()

    @action
    def quit(self):
        """ Quit the player (same as `stop`) """
        self._stop_torrent()
        if not self._player:
            return None, 'No vlc instance is running'

        self._player.stop()
        self._player = None
        return self.status()

    @action
    def stop(self):
        """ Stop the application (same as `quit`) """
        return self.quit()

    @action
    def voldown(self, step=10.0):
        """ Volume down by (default: 10)% """
        if not self._player:
            return None, 'No vlc instance is running'
        return self.set_volume(self._player.audio_get_volume()-step)

    @action
    def volup(self, step=10.0):
        """ Volume up by (default: 10)% """
        if not self._player:
            return None, 'No vlc instance is running'
        return self.set_volume(self._player.audio_get_volume()+step)

    @action
    def set_volume(self, volume):
        """
        Set the volume

        :param volume: Volume value between 0 and 100
        :type volume: float
        """
        if not self._player:
            return None, 'No vlc instance is running'

        volume = max(0, min([100, volume]))
        self._player.audio_set_volume(volume)
        status = self.status().output
        status['volume'] = volume
        return status

    @action
    def seek(self, position):
        """
        Seek backward/forward by the specified number of seconds

        :param position: Number of seconds relative to the current cursor
        :type position: int
        """
        if not self._player:
            return None, 'No vlc instance is running'
        if not self._player.is_seekable():
            return None, 'The resource is not seekable'

        media = self._player.get_media()
        if not media:
            return None, 'No media loaded'

        pos = min(media.get_duration()/1000, max(0, position))
        self._player.set_time(int(pos*1000))
        return self.status()

    @action
    def back(self, offset=60.0):
        """ Back by (default: 60) seconds """
        if not self._player:
            return None, 'No vlc instance is running'

        media = self._player.get_media()
        if not media:
            return None, 'No media loaded'

        pos = max(0, (self._player.get_time()/1000)-offset)
        return self.seek(pos)

    @action
    def forward(self, offset=60.0):
        """ Forward by (default: 60) seconds """
        if not self._player:
            return None, 'No vlc instance is running'

        media = self._player.get_media()
        if not media:
            return None, 'No media loaded'

        pos = min(media.get_duration()/1000, (self._player.get_time()/1000)+offset)
        return self.seek(pos)

    @action
    def toggle_subtitles(self, visibile=None):
        """ Toggle the subtitles visibility """
        if not self._player:
            return None, 'No vlc instance is running'

        if self._player.video_get_spu_count() == 0:
            return None, 'The media file has no subtitles set'

        if self._player.video_get_spu() is None or \
                self._player.video_get_spu() == -1:
            self._player.video_set_spu(0)
        else:
            self._player.video_set_spu(-1)

    @action
    def toggle_fullscreen(self):
        """ Toggle the fullscreen mode """
        if not self._player:
            return None, 'No vlc instance is running'
        self._player.toggle_fullscreen()

    @action
    def set_fullscreen(self, fullscreen=True):
        """ Set fullscreen mode """
        if not self._player:
            return None, 'No vlc instance is running'
        self._player.set_fullscreen(fullscreen)

    @action
    def set_subtitles(self, filename, **args):
        """ Sets media subtitles from filename """
        if not self._player:
            return None, 'No vlc instance is running'
        if filename.startswith('file://'):
            filename = filename[len('file://'):]

        self._player.video_set_subtitle_file(filename)

    @action
    def remove_subtitles(self):
        """ Removes (hides) the subtitles """
        if not self._player:
            return None, 'No vlc instance is running'
        self._player.video_set_spu(-1)

    @action
    def is_playing(self):
        """
        :returns: True if it's playing, False otherwise
        """
        if not self._player:
            return False
        return self._player.is_playing()

    @action
    def load(self, resource, **args):
        """
        Load/queue a resource/video to the player
        """
        if not self._player:
            return self.play(resource, **args)
        self._player.set_media(resource)
        return self.status()

    @action
    def mute(self):
        """ Toggle mute state """
        if not self._player:
            return None, 'No vlc instance is running'
        self._player.audio_toggle_mute()

    @action
    def set_position(self, position):
        """
        Seek backward/forward to the specified absolute position (same as ``seek``)
        """
        return self.seek(position)

    @action
    def status(self):
        """
        Get the current player state.

        :returns: A dictionary containing the current state.

        Example::

            output = {
                "filename": "filename or stream URL",
                "state": "play"  # or "stop" or "pause"
            }
        """
        import vlc
        if not self._player:
            return {'state': PlayerState.STOP.value}

        status = {}
        vlc_state = self._player.get_state()

        if vlc_state == vlc.State.Playing:
            status['state'] = PlayerState.PLAY.value
        elif vlc_state == vlc.State.Paused:
            status['state'] = PlayerState.PAUSE.value
        else:
            status['state'] = PlayerState.STOP.value

        status['url'] = urllib.parse.unquote(self._player.get_media().get_mrl()) if self._player.get_media() else None
        status['position'] = float(self._player.get_time()/1000) if self._player.get_time() is not None else None

        media = self._player.get_media()
        status['duration'] = media.get_duration()/1000 if media and media.get_duration() is not None else None

        status['seekable'] = status['duration'] is not None
        status['fullscreen'] = self._player.get_fullscreen()
        status['mute'] = self._player.audio_get_mute()
        status['path'] = status['url']
        status['pause'] = status['state'] == PlayerState.PAUSE.value
        status['percent_pos'] = self._player.get_position()*100
        status['filename'] = self._filename
        status['title'] = self._title
        status['volume'] = self._player.audio_get_volume()
        status['volume_max'] = 100

        return status

    def on_stop(self, callback):
        self._on_stop_callbacks.append(callback)

    def _get_current_resource(self):
        if not self._player or not self._player.get_media():
            return
        return self._player.get_media().get_mrl()


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import base64
import datetime
import os

from platypush.plugins import action
from platypush.plugins.google import GooglePlugin
from platypush.plugins.calendar import CalendarInterface


class GoogleCalendarPlugin(GooglePlugin, CalendarInterface):
    """
    Google calendar plugin
    """

    scopes = ['https://www.googleapis.com/auth/calendar.readonly']

    def __init__(self, *args, **kwargs):
        super().__init__(scopes=self.scopes, *args, **kwargs)


    @action
    def get_upcoming_events(self, max_results=10):
        """
        Get the upcoming events. See
        :func:`~platypush.plugins.calendar.CalendarPlugin.get_upcoming_events`.
        """

        now = datetime.datetime.utcnow().isoformat() + 'Z'
        service = self.get_service('calendar', 'v3')
        result = service.events().list(calendarId='primary', timeMin=now,
                                       maxResults=max_results, singleEvents=True,
                                       orderBy='startTime').execute()

        events = result.get('items', [])
        return events


# vim:sw=4:ts=4:et:




import enum
import logging
import time

import croniter

from threading import Thread

from platypush.procedure import Procedure

logger = logging.getLogger(__name__)


class CronjobState(enum.IntEnum):
    IDLE = 0
    WAIT = 1
    RUNNING = 2
    DONE = 3
    ERROR = 4


class Cronjob(Thread):
    def __init__(self, name, cron_expression, actions):
        super().__init__()
        self.cron_expression = cron_expression
        self.name = name
        self.state = CronjobState.IDLE
        self.actions = Procedure.build(name=name + '__Cron', _async=False,
                                       requests=actions)

    def run(self):
        self.state = CronjobState.WAIT
        self.wait()
        self.state = CronjobState.RUNNING

        try:
            logger.info('Running cronjob {}'.format(self.name))
            context = {}
            response = self.actions.execute(_async=False, **context)
            logger.info('Response from cronjob {}: {}'.format(self.name, response))
            self.state = CronjobState.DONE
        except Exception as e:
            logger.exception(e)
            self.state = CronjobState.ERROR

    def wait(self):
        now = int(time.time())
        cron = croniter.croniter(self.cron_expression, now)
        next_run = int(cron.get_next())
        time.sleep(next_run - now)

    def should_run(self):
        now = int(time.time())
        cron = croniter.croniter(self.cron_expression, now)
        next_run = int(cron.get_next())
        return now == next_run


class CronScheduler(Thread):
    def __init__(self, jobs):
        super().__init__()
        self.jobs_config = jobs
        self._jobs = {}
        logger.info('Cron scheduler initialized with {} jobs'.
                    format(len(self.jobs_config.keys())))

    def _get_job(self, name, config):
        job = self._jobs.get(name)
        if job and job.state not in [CronjobState.DONE, CronjobState.ERROR]:
            return job

        self._jobs[name] = Cronjob(name=name, cron_expression=config['cron_expression'],
                                   actions=config['actions'])

        return self._jobs[name]

    def run(self):
        logger.info('Running cron scheduler')

        while True:
            for (job_name, job_config) in self.jobs_config.items():
                job = self._get_job(name=job_name, config=job_config)
                if job.state == CronjobState.IDLE:
                    job.start()

            time.sleep(0.5)


# vim:sw=4:ts=4:et:



import json

from pynodered import node_red
from platypush.context import get_plugin


# noinspection PyUnusedLocal
@node_red(name='run', title='run', category='platypush', description='Run a platypush action')
def run(node, msg):
    msg = msg['payload']
    if isinstance(msg, bytes):
        msg = msg.decode()
    if isinstance(msg, str):
        msg = json.loads(msg)

    assert isinstance(msg, dict) and 'action' in msg

    if 'type' not in msg:
        msg['type'] = 'request'

    plugin_name = '.'.join(msg['action'].split('.')[:-1])
    action_name = msg['action'].split('.')[-1]
    plugin = get_plugin(plugin_name)
    action = getattr(plugin, action_name)
    args = msg.get('args', {})

    response = action(**args)
    if response.errors:
        raise response.errors[0]

    msg['payload'] = response.output
    return msg


# vim:sw=4:ts=4:et:



import enum
import logging
import re

from queue import LifoQueue
from ..config import Config
from ..message.request import Request
from ..message.response import Response

logger = logging.getLogger(__name__)


class Statement(enum.Enum):
    BREAK = 'break'
    CONTINUE = 'continue'
    RETURN = 'return'


class Procedure(object):
    """ Procedure class. A procedure is a pre-configured list of requests """

    def __init__(self, name, _async, requests, args=None, backend=None):
        """
        Params:
            name     -- Procedure name
            _async    -- Whether the actions in the procedure are supposed to
                        be executed sequentially or in parallel (True or False)
            requests -- List of platypush.message.request.Request objects
        """

        self.name = name
        self._async = _async
        self.requests = requests
        self.backend = backend
        self.args = args or {}
        self._should_return = False

        for req in requests:
            req.backend = self.backend

    @classmethod
    def build(cls, name, _async, requests, args=None, backend=None, id=None, procedure_class=None, **kwargs):
        reqs = []
        for_count = 0
        while_count = 0
        if_count = 0
        if_config = LifoQueue()
        procedure_class = procedure_class or cls
        key = None

        for request_config in requests:
            # Check if it's a break/continue/return statement
            if isinstance(request_config, str):
                reqs.append(Statement(request_config))
                continue

            # Check if this request is an if-else
            if len(request_config.keys()) >= 1:
                key = list(request_config.keys())[0]
                m = re.match('\s*(if)\s+\${(.*)}\s*', key)

                if m:
                    if_count += 1
                    if_name = '{}__if_{}'.format(name, if_count)
                    condition = m.group(2)

                    if_config.put({
                        'name': if_name,
                        '_async': False,
                        'requests': request_config[key],
                        'condition': condition,
                        'else_branch': [],
                        'backend': backend,
                        'id': id,
                    })

                    continue

                if key == 'else':
                    if if_config.empty():
                        raise RuntimeError('else statement with no ' +
                                           'associated if in {}'.format(name))

                    conf = if_config.get()
                    conf['else_branch'] = request_config[key]
                    if_config.put(conf)

            if not if_config.empty():
                reqs.append(IfProcedure.build(**(if_config.get())))
                if key == 'else':
                    continue

            # Check if this request is a for loop
            if len(request_config.keys()) == 1:
                key = list(request_config.keys())[0]
                m = re.match('\s*(fork?)\s+([\w\d_]+)\s+in\s+(.*)\s*', key)

                if m:
                    for_count += 1
                    loop_name = '{}__for_{}'.format(name, for_count)

                    # A 'for' loop is synchronous. Declare a 'fork' loop if you
                    # want to process the elements in the iterable in parallel
                    _async = True if m.group(1) == 'fork' else False
                    iterator_name = m.group(2)
                    iterable = m.group(3)

                    loop = ForProcedure.build(name=loop_name, _async=_async,
                                              requests=request_config[key],
                                              backend=backend, id=id,
                                              iterator_name=iterator_name,
                                              iterable=iterable)

                    reqs.append(loop)
                    continue

            # Check if this request is a while loop
            if len(request_config.keys()) == 1:
                key = list(request_config.keys())[0]
                m = re.match('\s*while\s+\${(.*)}\s*', key)

                if m:
                    while_count += 1
                    loop_name = '{}__while_{}'.format(name, while_count)
                    condition = m.group(1).strip()

                    loop = WhileProcedure.build(name=loop_name, _async=False,
                                                requests=request_config[key],
                                                condition=condition,
                                                backend=backend, id=id)

                    reqs.append(loop)
                    continue

            request_config['origin'] = Config.get('device_id')
            request_config['id'] = id
            if 'target' not in request_config:
                request_config['target'] = request_config['origin']

            request = Request.build(request_config)
            reqs.append(request)

        while not if_config.empty():
            pending_if = if_config.get()
            reqs.append(IfProcedure.build(**pending_if))

        return procedure_class(name=name, _async=_async, requests=reqs, args=args, backend=backend, **kwargs)

    @staticmethod
    def _find_nearest_loop(stack):
        for proc in stack[::-1]:
            if isinstance(proc, LoopProcedure):
                return proc

        raise AssertionError('break/continue statement found outside of a loop')

    def execute(self, n_tries=1, __stack__=None, **context):
        """
        Execute the requests in the procedure
        Params:
            n_tries -- Number of tries in case of failure before raising a RuntimeError
        """
        if not __stack__:
            __stack__ = [self]
        else:
            __stack__.append(self)

        if self.args:
            args = self.args.copy()
            for k, v in args.items():
                v = Request.expand_value_from_context(v, **context)
                args[k] = v
                context[k] = v
            logger.info('Executing procedure {} with arguments {}'.format(self.name, args))
        else:
            logger.info('Executing procedure {}'.format(self.name))

        response = Response()
        token = Config.get('token')

        for request in self.requests:
            if isinstance(request, Statement):
                if request == Statement.RETURN:
                    self._should_return = True
                    for proc in __stack__:
                        proc._should_return = True
                    break

                if request in [Statement.BREAK, Statement.CONTINUE]:
                    loop = self._find_nearest_loop(__stack__)
                    if request == Statement.BREAK:
                        loop._should_break = True
                    else:
                        loop._should_continue = True
                    break

            if isinstance(self, LoopProcedure):
                if self._should_continue or self._should_break:
                    if self._should_continue:
                        # noinspection PyAttributeOutsideInit
                        self._should_continue = False

                    break

            if token:
                request.token = token

            context['_async'] = self._async
            context['n_tries'] = n_tries
            response = request.execute(__stack__=__stack__, **context)

            if not self._async and response:
                if isinstance(response.output, dict):
                    for k, v in response.output.items():
                        context[k] = v

                context['output'] = response.output
                context['errors'] = response.errors

            if self._should_return:
                break

        return response or Response()


class LoopProcedure(Procedure):
    """
    Base class while and for/fork loops.
    """

    def __init__(self, name, requests, _async=False, args=None, backend=None):
        super(). __init__(name=name, _async=_async, requests=requests, args=args, backend=backend)
        self._should_break = False
        self._should_continue = False


class ForProcedure(LoopProcedure):
    """
    Models a loop procedure, which expresses a construct similar to a
    for loop in a programming language. The 'for' keyword implies a synchronous
    loop, i.e. the nested actions will be executed in sequence. Use 'fork'
    instead of 'for' if you want to run the actions in parallel.

    Example::

        procedure.sync.process_results:
            - action: http.get
              args:
                  url: https://some-service/some/json/endpoint
                  # Example response: { "results": [ {"id":1, "name":"foo"}, {"id":2,"name":"bar"} ]}
            - for result in ${results}:
                - action: some.custom.action
                  args:
                      id: ${result['id']}
                      name: ${result['name']}

    """

    def __init__(self, name, iterator_name, iterable, requests, _async=False, args=None, backend=None):
        super(). __init__(name=name, _async=_async, requests=requests, args=args, backend=backend)
        self.iterator_name = iterator_name
        self.iterable = iterable

    def execute(self, _async=None, **context):
        # noinspection PyBroadException
        try:
            iterable = eval(self.iterable)
            assert hasattr(iterable, '__iter__'), 'Object of type {} is not iterable: {}'.\
                format(type(iterable), iterable)
        except:
            iterable = Request.expand_value_from_context(self.iterable, **context)

        response = Response()

        # noinspection DuplicatedCode
        for item in iterable:
            if self._should_return:
                logger.info('Returning from {}'.format(self.name))
                break

            if self._should_continue:
                self._should_continue = False
                logger.info('Continuing loop {}'.format(self.name))
                continue

            if self._should_break:
                self._should_break = False
                logger.info('Breaking loop {}'.format(self.name))
                break

            context[self.iterator_name] = item
            response = super().execute(**context)

        return response


class WhileProcedure(LoopProcedure):
    """
    Models a while loop procedure.

    Example::

        procedure.process_results:
            - action: http.get
              args:
                  url: https://some-service/some/json/endpoint
                  # Example response: {"id":1, "name":"foo"}}

            - while ${output}:
                - action: some.custom.action
                  args:
                      id: ${id}
                      name: ${name}

                - action: http.get
                  args:
                      url: https://some-service/some/json/endpoint
                      # Example response: {"id":1, "name":"foo"}}

    """

    def __init__(self, name, condition, requests, _async=False, args=None, backend=None):
        super(). __init__(name=name, _async=_async, requests=requests, args=args, backend=backend)
        self.condition = condition

    @staticmethod
    def _get_context(**context):
        for (k, v) in context.items():
            # noinspection PyBroadException
            try:
                context[k] = eval(v)
            except:
                if isinstance(v, str):
                    # noinspection PyBroadException
                    try:
                        context[k] = eval('"{}"'.format(re.sub('(^|[^\\\])"', '\1\\"', v)))
                    except:
                        pass

        return context

    # noinspection DuplicatedCode,PyBroadException
    def execute(self, _async=None, **context):
        response = Response()
        context = self._get_context(**context)
        for k, v in context.items():
            try:
                exec('{}={}'.format(k, v))
            except:
                pass

        while True:
            condition_true = eval(self.condition)
            if not condition_true:
                break

            if self._should_return:
                logger.info('Returning from {}'.format(self.name))
                break

            if self._should_continue:
                self._should_continue = False
                logger.info('Continuing loop {}'.format(self.name))
                continue

            if self._should_break:
                self._should_break = False
                logger.info('Breaking loop {}'.format(self.name))
                break

            response = super().execute(**context)

            if response.output:
                if isinstance(response.output, dict):
                    new_context = self._get_context(**response.output)
                    for k, v in new_context.items():
                        try:
                            exec('{}={}'.format(k, v))
                        except:
                            pass

        return response


# noinspection PyBroadException
class IfProcedure(Procedure):
    """
    Models an if-else construct.

    Example:

        procedure.sync.process_results:
            -
                action: http.get
                args:
                    url: https://some-service/some/json/endpoint
                    # Example response: { "sensors": [ {"temperature": 18 } ] }
            -
                if ${sensors['temperature'] < 20}:
                    -
                        action: shell.exec
                        args:
                            cmd: '/path/turn_on_heating.sh'
                else:
                    -
                        action: shell.exec
                        args:
                            cmd: '/path/turn_off_heating.sh'
    """

    def __init__(self, name, condition, requests, else_branch=None, args=None, backend=None, id=None, **kwargs):
        kwargs['_async'] = False
        self.condition = condition
        self.else_branch = else_branch
        reqs = []

        for req in requests:
            if isinstance(req, dict):
                req['origin'] = Config.get('device_id')
                req['id'] = id
                if 'target' not in req:
                    req['target'] = req['origin']

                req = Request.build(req)

            reqs.append(req)

        super(). __init__(name=name, requests=reqs, args=args, backend=backend, **kwargs)

    @classmethod
    def build(cls, name, condition, requests, else_branch=None, args=None, backend=None, id=None, **kwargs):
        kwargs['_async'] = False
        if else_branch:
            else_branch = super().build(name=name+'__else', requests=else_branch,
                                        args=args, backend=backend, id=id,
                                        procedure_class=Procedure, **kwargs)

        return super().build(name=name, condition=condition, requests=requests,
                             else_branch=else_branch, args=args, backend=backend, id=id,
                             **kwargs)

    def execute(self, **context):
        for (k, v) in context.items():
            try:
                exec('{}={}'.format(k, v))
            except:
                if isinstance(v, str):
                    try:
                        exec('{}="{}"'.format(k, re.sub('(^|[^\\\])"', '\1\\"', v)))
                    except:
                        pass

        condition_true = eval(self.condition)
        response = Response()

        if condition_true:
            response = super().execute(**context)
        elif self.else_branch:
            response = self.else_branch.execute(**context)

        return response


# vim:sw=4:ts=4:et:



from datetime import datetime
from typing import Optional, List, Union

from platypush.message.response import Response


class SystemResponse(Response):
    pass


class CpuResponse(SystemResponse):
    pass


class MemoryResponse(SystemResponse):
    pass


class DiskResponse(SystemResponse):
    pass


class NetworkResponse(SystemResponse):
    pass


class SensorResponse(SystemResponse):
    pass


class CpuInfoResponse(CpuResponse):
    def __init__(self,
                 arch: str,
                 bits: int,
                 count: int,
                 vendor_id: str,
                 brand: str,
                 hz_advertised: int,
                 hz_actual: int,
                 model: int,
                 flags: List[str],
                 family: Optional[int],
                 stepping: Optional[int],
                 l1_instruction_cache_size: Optional[Union[int, str]],
                 l1_data_cache_size: Optional[Union[int, str]],
                 l2_cache_size: Optional[Union[int, str]],
                 l3_cache_size: Optional[Union[int, str]],
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'arch': arch,
                'bits': bits,
                'count': count,
                'vendor_id': vendor_id,
                'brand': brand,
                'hz_advertised': hz_advertised,
                'hz_actual': hz_actual,
                'stepping': stepping,
                'model': model,
                'family': family,
                'flags': flags,
                'l1_instruction_cache_size': l1_instruction_cache_size,
                'l1_data_cache_size': l1_data_cache_size,
                'l2_cache_size': l2_cache_size,
                'l3_cache_size': l3_cache_size,
            }, **kwargs
        )


class CpuTimesResponse(CpuResponse):
    def __init__(self,
                 user: float,
                 nice: float,
                 system: float,
                 idle: float,
                 iowait: float,
                 irq: float,
                 softirq: float,
                 steal: float,
                 guest: float,
                 guest_nice: float,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'user': user,
                'nice': nice,
                'system': system,
                'idle': idle,
                'iowait': iowait,
                'irq': irq,
                'softirq': softirq,
                'steal': steal,
                'guest': guest,
                'guest_nice': guest_nice,
            }, **kwargs
        )


class CpuStatsResponse(CpuResponse):
    def __init__(self,
                 ctx_switches: int,
                 interrupts: int,
                 soft_interrupts: int,
                 syscalls: int,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'ctx_switches': ctx_switches,
                'interrupts': interrupts,
                'soft_interrupts': soft_interrupts,
                'syscalls': syscalls,
            }, **kwargs
        )


class CpuFrequencyResponse(CpuResponse):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 min: int,
                 max: int,
                 current: int,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'min': min,
                'max': max,
                'current': current,
            }, **kwargs
        )


class VirtualMemoryUsageResponse(MemoryResponse):
    def __init__(self,
                 total: int,
                 available: int,
                 percent: float,
                 used: int,
                 free: int,
                 active: int,
                 inactive: int,
                 buffers: int,
                 cached: int,
                 shared: int,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'total': total,
                'available': available,
                'percent': percent,
                'used': used,
                'free': free,
                'active': active,
                'inactive': inactive,
                'buffers': buffers,
                'cached': cached,
                'shared': shared,
            }, **kwargs
        )


class SwapMemoryUsageResponse(MemoryResponse):
    def __init__(self,
                 total: int,
                 percent: float,
                 used: int,
                 free: int,
                 sin: int,
                 sout: int,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'total': total,
                'percent': percent,
                'used': used,
                'free': free,
                'sin': sin,
                'sout': sout,
            }, **kwargs
        )


class DiskPartitionResponse(DiskResponse):
    def __init__(self,
                 device: str,
                 mount_point: str,
                 fstype: Optional[str] = None,
                 opts: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                 'device': device,
                 'mount_point': mount_point,
                 'fstype': fstype,
                 'opts': opts,
            }, **kwargs
        )


class DiskUsageResponse(DiskResponse):
    def __init__(self,
                 path: str,
                 total: int,
                 used: int,
                 free: int,
                 percent: float,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                 'path': path,
                 'total': total,
                 'used': used,
                 'free': free,
                 'percent': percent,
            }, **kwargs
        )


class DiskIoCountersResponse(DiskResponse):
    def __init__(self,
                 read_count: int,
                 write_count: int,
                 read_bytes: int,
                 write_bytes: int,
                 read_time: int,
                 write_time: int,
                 read_merged_count: int,
                 write_merged_count: int,
                 busy_time: int,
                 disk: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'read_count': read_count,
                'write_count': write_count,
                'read_bytes': read_bytes,
                'write_bytes': write_bytes,
                'read_time': read_time,
                'write_time': write_time,
                'read_merged_count': read_merged_count,
                'write_merged_count': write_merged_count,
                'busy_time': busy_time,
                'disk': disk,
            }, **kwargs
        )


class NetworkIoCountersResponse(NetworkResponse):
    def __init__(self,
                 bytes_sent: int,
                 bytes_recv: int,
                 packets_sent: int,
                 packets_recv: int,
                 errin: int,
                 errout: int,
                 dropin: int,
                 dropout: int,
                 nic: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'bytes_sent': bytes_sent,
                'bytes_recv': bytes_recv,
                'packets_sent': packets_sent,
                'packets_recv': packets_recv,
                'errin': errin,
                'errout': errout,
                'dropin': dropin,
                'dropout': dropout,
                'nic': nic,
            }, **kwargs
        )


class NetworkConnectionResponse(NetworkResponse):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 fd: int,
                 family: str,
                 type: str,
                 local_address: str,
                 local_port: int,
                 remote_address: str,
                 remote_port: int,
                 status: str,
                 pid: int,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'fd': fd,
                'family': family,
                'type': type,
                'local_address': local_address,
                'local_port': local_port,
                'remote_address': remote_address,
                'remote_port': remote_port,
                'status': status,
                'pid': pid,
            }, **kwargs
        )


class NetworkAddressResponse(NetworkResponse):
    def __init__(self,
                 nic: str,
                 ipv4_address: Optional[str] = None,
                 ipv4_netmask: Optional[str] = None,
                 ipv4_broadcast: Optional[str] = None,
                 ipv6_address: Optional[str] = None,
                 ipv6_netmask: Optional[str] = None,
                 ipv6_broadcast: Optional[str] = None,
                 mac_address: Optional[str] = None,
                 mac_broadcast: Optional[str] = None,
                 ptp: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'nic': nic,
                'ipv4_address': ipv4_address,
                'ipv4_netmask': ipv4_netmask,
                'ipv4_broadcast': ipv4_broadcast,
                'ipv6_address': ipv6_address,
                'ipv6_netmask': ipv6_netmask,
                'ipv6_broadcast': ipv6_broadcast,
                'mac_address': mac_address,
                'mac_broadcast': mac_broadcast,
                'ptp': ptp,
            }, **kwargs
        )


class NetworkInterfaceStatsResponse(NetworkResponse):
    def __init__(self,
                 nic: str,
                 is_up: bool,
                 duplex: str,
                 speed: int,
                 mtu: int,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'nic': nic,
                'is_up': is_up,
                'duplex': duplex,
                'speed': speed,
                'mtu': mtu,
            }, **kwargs
        )


class SensorTemperatureResponse(SensorResponse):
    def __init__(self,
                 name: str,
                 current: float,
                 high: Optional[float] = None,
                 critical: Optional[float] = None,
                 label: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'name': name,
                'current': current,
                'high': high,
                'critical': critical,
                'label': label,
            }, **kwargs
        )


class SensorFanResponse(SensorResponse):
    def __init__(self,
                 name: str,
                 current: int,
                 label: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'name': name,
                'current': current,
                'label': label,
            }, **kwargs
        )


class SensorBatteryResponse(SensorResponse):
    def __init__(self,
                 percent: float,
                 secs_left: int,
                 power_plugged: bool,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'percent': percent,
                'secs_left': secs_left,
                'power_plugged': power_plugged,
            }, **kwargs
        )


class ConnectUserResponse(SystemResponse):
    def __init__(self,
                 name: str,
                 terminal: str,
                 host: str,
                 started: datetime,
                 pid: Optional[int] = None,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'name': name,
                'terminal': terminal,
                'host': host,
                'started': started,
                'pid': pid,
            }, **kwargs
        )


class ProcessResponse(SystemResponse):
    def __init__(self,
                 pid: int,
                 name: str,
                 started: datetime,
                 ppid: Optional[int],
                 children: Optional[List[int]] = None,
                 exe: Optional[List[str]] = None,
                 status: Optional[str] = None,
                 username: Optional[str] = None,
                 terminal: Optional[str] = None,
                 cpu_user_time: Optional[float] = None,
                 cpu_system_time: Optional[float] = None,
                 cpu_children_user_time: Optional[float] = None,
                 cpu_children_system_time: Optional[float] = None,
                 mem_rss: Optional[int] = None,
                 mem_vms: Optional[int] = None,
                 mem_shared: Optional[int] = None,
                 mem_text: Optional[int] = None,
                 mem_data: Optional[int] = None,
                 mem_lib: Optional[int] = None,
                 mem_dirty: Optional[int] = None,
                 mem_percent: Optional[float] = None,
                 *args, **kwargs):
        super().__init__(
            *args, output={
                'pid': pid,
                'name': name,
                'started': started,
                'ppid': ppid,
                'exe': exe,
                'status': status,
                'username': username,
                'terminal': terminal,
                'cpu_user_time': cpu_user_time,
                'cpu_system_time': cpu_system_time,
                'cpu_children_user_time': cpu_children_user_time,
                'cpu_children_system_time': cpu_children_system_time,
                'mem_rss': mem_rss,
                'mem_vms': mem_vms,
                'mem_shared': mem_shared,
                'mem_text': mem_text,
                'mem_data': mem_data,
                'mem_lib': mem_lib,
                'mem_dirty': mem_dirty,
                'mem_percent': mem_percent,
                'children': children or [],
            }, **kwargs
        )


class SystemResponseList(SystemResponse):
    def __init__(self, responses: List[SystemResponse], *args, **kwargs):
        super().__init__(output=[r.output for r in responses], *args, **kwargs)


class CpuResponseList(CpuResponse, SystemResponseList):
    def __init__(self, responses: List[CpuResponse], *args, **kwargs):
        super().__init__(responses=responses, *args, **kwargs)


class DiskResponseList(DiskResponse, SystemResponseList):
    def __init__(self, responses: List[DiskResponse], *args, **kwargs):
        super().__init__(responses=responses, *args, **kwargs)


class NetworkResponseList(NetworkResponse, SystemResponseList):
    def __init__(self, responses: List[NetworkResponse], *args, **kwargs):
        super().__init__(responses=responses, *args, **kwargs)


class SensorResponseList(SensorResponse, SystemResponseList):
    def __init__(self, responses: List[SensorResponse], *args, **kwargs):
        super().__init__(responses=responses, *args, **kwargs)


class ConnectedUserResponseList(SystemResponseList):
    def __init__(self, responses: List[ConnectUserResponse], *args, **kwargs):
        super().__init__(responses=responses, *args, **kwargs)


class ProcessResponseList(SystemResponseList):
    def __init__(self, responses: List[ProcessResponse], *args, **kwargs):
        super().__init__(responses=responses, *args, **kwargs)


# vim:sw=4:ts=4:et:



from platypush.plugins import action
from platypush.plugins.gpio.sensor import GpioSensorPlugin


class GpioSensorEnvirophatPlugin(GpioSensorPlugin):
    """
    Plugin to interact with a `Pimoroni enviropHAT <https://shop.pimoroni.com/products/enviro-phat>`_ device.
    You can use an enviropHAT device to read e.g. temperature, pressure, altitude, accelerometer, magnetometer and
    luminosity data, plus control the status of its RGB LEDs.

    Requires:

        * ``envirophat`` (``pip install envirophat``)
    """

    @action
    def get_measurement(self, qnh=1020.0):
        """
        :param: qnh: Local value for atmospheric pressure adjusted to sea level (default: 1020)
        :type qnh: float

        :returns: dict. Example:

        .. code-block:: python

           output = {
               "temperature": 21.0,   # Celsius
               "pressure": 101555.08, # pascals
               "altitude": 10,        # meters
               "luminosity": 426,     # lumens

               # Measurements from the custom analog channels
               "analog": [0.513, 0.519, 0.531, 0.528],

               "accelerometer": {
                   "x": -0.000915,
                   "y": 0.0760,
                   "z": 1.026733
               },
               "magnetometer": {
                   "x": -2297,
                   "y": 1226,
                   "z": -7023
               },
           }

        """

        import envirophat

        ret = {}
        weather = envirophat.weather
        light = envirophat.light
        accelerometer = envirophat.motion.accelerometer()
        magnetometer = envirophat.motion.magnetometer()
        leds = envirophat.leds
        analog = envirophat.analog

        weather.update()

        ret['temperature'] = weather.temperature()
        ret['pressure'] = weather.pressure()
        ret['altitude'] = weather.altitude(qnh=qnh)
        ret['luminosity'] = light.light()
        ret['accelerometer'] = {v: getattr(accelerometer, v) for v in ['x', 'y', 'z']}
        ret['magnetometer'] = {v: getattr(magnetometer, v) for v in ['x', 'y', 'z']}
        ret['analog'] = list(analog.read_all())
        ret['leds'] = leds.is_on()

        return ret


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class GPSEvent(Event):
    """
    Generic class for GPS events
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class GPSVersionEvent(GPSEvent):
    """
    Event usually triggered on startup or reconnection, when the GPS device advertises its version parameters
    """

    def __init__(self, release=None, rev=None, proto_major=None, proto_minor=None, *args, **kwargs):
        super().__init__(release=release, rev=rev, proto_major=proto_major, proto_minor=proto_minor, *args, **kwargs)


class GPSDeviceEvent(GPSEvent):
    """
    Event triggered when a new GPS device is connected or reconfigured
    """

    def __init__(self, path, activated=None, native=False, bps=None, parity=None, stopbits=None,
                 cycle=None, driver=None, *args, **kwargs):
        super().__init__(*args, path=path, activated=activated, native=native, bps=bps, parity=parity,
                         stopbits=stopbits, cycle=cycle, driver=driver, **kwargs)


class GPSUpdateEvent(GPSEvent):
    """
    Event triggered upon GPS status update
    """

    def __init__(self, device=None, latitude=None, longitude=None, altitude=None, mode=None, epv=None, eph=None,
                 sep=None, *args, **kwargs):
        super().__init__(*args, device=device, latitude=latitude, longitude=longitude, altitude=altitude,
                         mode=mode, epv=epv, eph=eph, sep=sep, **kwargs)


# vim:sw=4:ts=4:et:



import json
import re
import threading
import time

from platypush.context import get_bus
from platypush.plugins import action
from platypush.plugins.media import MediaPlugin, PlayerState
from platypush.message.event.media import MediaPlayEvent, MediaPauseEvent, MediaStopEvent, \
    MediaSeekEvent, MediaVolumeChangedEvent, NewPlayingMediaEvent


# noinspection PyUnusedLocal
class MediaKodiPlugin(MediaPlugin):
    """
    Plugin to interact with a Kodi media player instance

    Requires:

        * **kodi-json** (``pip install kodi-json``)
        * **websocket-client** (``pip install websocket-client``), optional, for player events support
    """

    def __init__(self, host, http_port=8080, websocket_port=9090, username=None, password=None, **kwargs):
        """
        :param host: Kodi host name or IP
        :type host: str

        :param http_port: Kodi JSON RPC web port. Remember to enable "Allow remote control via HTTP"
            in Kodi service settings -> advanced configuration and "Allow remote control from applications"
            on this system and, optionally, on other systems if the Kodi server is on another machine
        :type http_port: int

        :param websocket_port: Kodi JSON RPC websocket port, used to receive player events
        :type websocket_port: int

        :param username: Kodi username (optional)
        :type username: str

        :param password: Kodi password (optional)
        :type password: str
        """

        super().__init__(**kwargs)

        self.host = host
        self.http_port = http_port
        self.websocket_port = websocket_port
        self.url = 'http://{}:{}/jsonrpc'.format(host, http_port)
        self.websocket_url = 'ws://{}:{}/jsonrpc'.format(host, websocket_port)
        self.username = username
        self.password = password
        self._ws = None
        threading.Thread(target=self._websocket_thread()).start()

    def _get_kodi(self):
        from kodijson import Kodi

        args = [self.url]
        if self.username:
            args += [self.username]
            if self.password:
                args += [self.password]

        return Kodi(*args)

    def _get_player_id(self):
        kodi = self._get_kodi()
        players = kodi.Player.GetActivePlayers().get('result', [])
        if not players:
            return None

        return players.pop().get('playerid')

    def _websocket_thread(self):
        """
        Initialize the websocket JSON RPC interface, if available, to receive player notifications
        """

        def thread_hndl():
            try:
                import websocket
            except ImportError:
                self.logger.warning('websocket-client is not installed, Kodi events will be disabled')
                return

            if not self._ws:
                self._ws = websocket.WebSocketApp(self.websocket_url,
                                                  on_message=self._on_ws_msg(),
                                                  on_error=self._on_ws_error(),
                                                  on_close=self._on_ws_close())

                self.logger.info('Kodi websocket interface for events started')
                self._ws.run_forever()

        return thread_hndl

    def _post_event(self, evt_type, **evt):
        bus = get_bus()
        bus.post(evt_type(player=self.host, plugin='media.kodi', **evt))

    def _on_ws_msg(self):
        def hndl(ws, msg):
            self.logger.info("Received Kodi message: {}".format(msg))
            msg = json.loads(msg)
            method = msg.get('method')

            if method == 'Player.OnPlay':
                item = msg.get('params', {}).get('data', {}).get('item', {})
                player = msg.get('params', {}).get('data', {}).get('player', {})
                self._post_event(MediaPlayEvent, player_id=player.get('playerid'),
                                 title=item.get('title'), media_type=item.get('type'))
            elif method == 'Player.OnPause':
                item = msg.get('params', {}).get('data', {}).get('item', {})
                player = msg.get('params', {}).get('data', {}).get('player', {})
                self._post_event(MediaPauseEvent, player_id=player.get('playerid'),
                                 title=item.get('title'), media_type=item.get('type'))
            elif method == 'Player.OnStop':
                player = msg.get('params', {}).get('data', {}).get('player', {})
                self._post_event(MediaStopEvent, player_id=player.get('playerid'))
            elif method == 'Player.OnSeek':
                player = msg.get('params', {}).get('data', {}).get('player', {})
                position = self._time_obj_to_pos(player.get('seekoffset'))
                self._post_event(MediaSeekEvent, position=position, player_id=player.get('playerid'))
            elif method == 'Application.OnVolumeChanged':
                volume = msg.get('params', {}).get('data', {}).get('volume')
                self._post_event(MediaVolumeChangedEvent, volume=volume)

        return hndl

    def _on_ws_error(self):
        def hndl(ws, error):
            self.logger.warning("Kodi websocket connection error: {}".format(error))
        return hndl

    def _on_ws_close(self):
        def hndl(ws):
            self._ws = None
            self.logger.warning("Kodi websocket connection closed")
            time.sleep(5)
            self._websocket_thread()

        return hndl

    def _build_result(self, result):
        status = self.status().output
        status['result'] = result.get('result')
        return status, result.get('error')

    @action
    def play(self, resource, *args, **kwargs):
        """
        Open and play the specified file or URL

        :param resource: URL or path to the media to be played
        """

        if resource.startswith('youtube:video:') \
                or resource.startswith('https://www.youtube.com/watch?v='):
            m1 = re.match('youtube:video:([^:?&]+)', resource)
            m2 = re.match('https://www.youtube.com/watch?v=([^:?&#/]+)', resource)
            youtube_id = None

            if m1:
                youtube_id = m1.group(1)
            elif m2:
                youtube_id = m2.group(1)

            if youtube_id:
                resource = 'plugin://plugin.video.youtube/?action=play_video&videoid=' + youtube_id

        if resource.startswith('file://'):
            resource = resource[7:]

        result = self._get_kodi().Player.Open(item={'file': resource})
        if self.volume:
            self.set_volume(volume=int(self.volume))

        return self._build_result(result)

    @action
    def pause(self, player_id=None, *args, **kwargs):
        """
        Play/pause the current media
        """

        if player_id is None:
            player_id = self._get_player_id()
        if player_id is None:
            return None, 'No active players found'

        result = self._get_kodi().Player.PlayPause(playerid=player_id)
        return self._build_result(result)

    @action
    def get_active_players(self):
        """
        Get the list of active players
        """

        result = self._get_kodi().Player.GetActivePlayers()
        return result.get('result'), result.get('error')

    @action
    def get_movies(self, *args, **kwargs):
        """
        Get the list of movies on the Kodi server
        """

        result = self._get_kodi().VideoLibrary.GetMovies()
        return result.get('result'), result.get('error')

    @action
    def stop(self, player_id=None, *args, **kwargs):
        """
        Stop the current media
        """

        if player_id is None:
            player_id = self._get_player_id()
        if player_id is None:
            return None, 'No active players found'

        result = self._get_kodi().Player.Stop(playerid=player_id)
        return self._build_result(result)

    @action
    def notify(self, title, message, *args, **kwargs):
        """
        Send a notification to the Kodi UI
        """

        result = self._get_kodi().GUI.ShowNotification(title=title, message=message)
        return result.get('result'), result.get('error')

    @action
    def left(self, *args, **kwargs):
        """
        Simulate a left input event
        """

        result = self._get_kodi().Input.Left()
        return result.get('result'), result.get('error')

    @action
    def right(self, *args, **kwargs):
        """
        Simulate a right input event
        """

        result = self._get_kodi().Input.Right()
        return result.get('result'), result.get('error')

    @action
    def up(self, *args, **kwargs):
        """
        Simulate an up input event
        """

        result = self._get_kodi().Input.Up()
        return result.get('result'), result.get('error')

    @action
    def down(self, *args, **kwargs):
        """
        Simulate a down input event
        """

        result = self._get_kodi().Input.Down()
        return result.get('result'), result.get('error')

    @action
    def back_btn(self, *args, **kwargs):
        """
        Simulate a back input event
        """

        result = self._get_kodi().Input.Back()
        return result.get('result'), result.get('error')

    @action
    def select(self, *args, **kwargs):
        """
        Simulate a select input event
        """

        result = self._get_kodi().Input.Select()
        return result.get('result'), result.get('error')

    @action
    def send_text(self, text, *args, **kwargs):
        """
        Simulate a send_text input event

        :param text: Text to send
        :type text: str
        """

        result = self._get_kodi().Input.SendText(text=text)
        return result.get('result'), result.get('error')

    @action
    def get_volume(self, *args, **kwargs):
        result = self._get_kodi().Application.GetProperties(
            properties=['volume'])

        return result.get('result'), result.get('error')

    @action
    def volup(self, step=10.0, *args, **kwargs):
        """ Volume up (default: +10%) """
        volume = self._get_kodi().Application.GetProperties(
            properties=['volume']).get('result', {}).get('volume')

        result = self._get_kodi().Application.SetVolume(volume=int(min(volume+step, 100)))
        return self._build_result(result)

    @action
    def voldown(self, step=10.0, *args, **kwargs):
        """ Volume down (default: -10%) """
        volume = self._get_kodi().Application.GetProperties(
            properties=['volume']).get('result', {}).get('volume')

        result = self._get_kodi().Application.SetVolume(volume=int(max(volume-step, 0)))
        return self._build_result(result)

    @action
    def set_volume(self, volume, *args, **kwargs):
        """
        Set the application volume

        :param volume: Volume to set between 0 and 100
        :type volume: int
        """

        result = self._get_kodi().Application.SetVolume(volume=int(volume))
        return self._build_result(result)

    @action
    def mute(self, *args, **kwargs):
        """
        Mute/unmute the application
        """

        muted = self._get_kodi().Application.GetProperties(
            properties=['muted']).get('result', {}).get('muted')

        result = self._get_kodi().Application.SetMute(mute=(not muted))
        return self._build_result(result)

    @action
    def is_muted(self, *args, **kwargs):
        """
        Return the muted status of the application
        """

        result = self._get_kodi().Application.GetProperties(properties=['muted'])
        return result.get('result')

    @action
    def scan_video_library(self, *args, **kwargs):
        """
        Scan the video library
        """

        result = self._get_kodi().VideoLibrary.Scan()
        return result.get('result'), result.get('error')

    @action
    def scan_audio_library(self, *args, **kwargs):
        """
        Scan the audio library
        """

        result = self._get_kodi().AudioLibrary.Scan()
        return result.get('result'), result.get('error')

    @action
    def clean_video_library(self, *args, **kwargs):
        """
        Clean the video library
        """

        result = self._get_kodi().VideoLibrary.Clean()
        return result.get('result'), result.get('error')

    @action
    def clean_audio_library(self, *args, **kwargs):
        """
        Clean the audio library
        """

        result = self._get_kodi().AudioLibrary.Clean()
        return result.get('result'), result.get('error')

    @action
    def quit(self, *args, **kwargs):
        """
        Quit the application
        """

        result = self._get_kodi().Application.Quit()
        return result.get('result'), result.get('error')

    @action
    def get_songs(self, *args, **kwargs):
        """
        Get the list of songs in the audio library
        """

        result = self._get_kodi().Application.GetSongs()
        return result.get('result'), result.get('error')

    @action
    def get_artists(self, *args, **kwargs):
        """
        Get the list of artists in the audio library
        """

        result = self._get_kodi().Application.GetArtists()
        return result.get('result'), result.get('error')

    @action
    def get_albums(self, *args, **kwargs):
        """
        Get the list of albums in the audio library
        """

        result = self._get_kodi().Application.GetAlbums()
        return result.get('result'), result.get('error')

    @action
    def fullscreen(self, *args, **kwargs):
        """
        Set/unset fullscreen mode
        """

        fullscreen = self._get_kodi().GUI.GetProperties(
            properties=['fullscreen']).get('result', {}).get('fullscreen')

        result = self._get_kodi().GUI.SetFullscreen(fullscreen=(not fullscreen))
        return result.get('result'), result.get('error')

    @action
    def shuffle(self, player_id=None, shuffle=None, *args, **kwargs):
        """
        Set/unset shuffle mode
        """

        if player_id is None:
            player_id = self._get_player_id()
        if player_id is None:
            return None, 'No active players found'

        if shuffle is None:
            shuffle = self._get_kodi().Player.GetProperties(
                playerid=player_id,
                properties=['shuffled']).get('result', {}).get('shuffled')

        result = self._get_kodi().Player.SetShuffle(
           playerid=player_id,  shuffle=(not shuffle))
        return result.get('result'), result.get('error')

    @action
    def repeat(self, player_id=None, repeat=None, *args, **kwargs):
        """
        Set/unset repeat mode
        """

        if player_id is None:
            player_id = self._get_player_id()
        if player_id is None:
            return None, 'No active players found'

        if repeat is None:
            repeat = self._get_kodi().Player.GetProperties(
                playerid=player_id,
                properties=['repeat']).get('result', {}).get('repeat')

        result = self._get_kodi().Player.SetRepeat(
            playerid=player_id,
            repeat='off' if repeat in ('one','all') else 'off')

        return result.get('result'), result.get('error')

    @staticmethod
    def _time_pos_to_obj(t):
        hours = int(t/3600)
        minutes = int((t - hours*3600)/60)
        seconds = t - hours*3600 - minutes*60
        milliseconds = t - int(t)

        return {
            'hours': hours,
            'minutes': minutes,
            'seconds': seconds,
            'milliseconds': milliseconds,
        }

    @staticmethod
    def _time_obj_to_pos(t):
        return t.get('hours', 0) * 3600 + t.get('minutes', 0) * 60 + \
               t.get('seconds', 0) + t.get('milliseconds', 0)/1000

    @action
    def seek(self, position, player_id=None, *args, **kwargs):
        """
        Move to the specified time position in seconds

        :param position: Seek time in seconds
        :type position: float
        """

        if player_id is None:
            player_id = self._get_player_id()
        if player_id is None:
            return None, 'No active players found'

        position = self._time_pos_to_obj(position)
        result = self._get_kodi().Player.Seek(playerid=player_id, value=position)
        return self._build_result(result)

    @action
    def set_position(self, position, player_id=None, *args, **kwargs):
        """
        Move to the specified time position in seconds

        :param position:  Seek time in seconds
        :type position: float
        """
        return self.seek(position=position, player_id=player_id, *args, **kwargs)

    @action
    def back(self, offset=60, player_id=None, *args, **kwargs):
        """
        Move the player execution backward by delta_seconds

        :param offset: Backward seek duration (default: 60 seconds)
        :type offset: float
        """

        if player_id is None:
            player_id = self._get_player_id()
        if player_id is None:
            return None, 'No active players found'

        position = self._get_kodi().Player.GetProperties(
            playerid=player_id, properties=['time']).get('result', {}).get('time', {})

        position = self._time_obj_to_pos(position)
        return self.seek(player_id=player_id, position=position)

    @action
    def forward(self, offset=60, player_id=None, *args, **kwargs):
        """
        Move the player execution forward by delta_seconds

        :param offset: Forward seek duration (default: 60 seconds)
        :type offset: float
        """

        if player_id is None:
            player_id = self._get_player_id()
        if player_id is None:
            return None, 'No active players found'

        position = self._get_kodi().Player.GetProperties(
            playerid=player_id, properties=['time']).get('result', {}).get('time', {})

        position = self._time_obj_to_pos(position)
        return self.seek(player_id=player_id, position=position)

    @action
    def status(self, player_id=None):
        media_props = {
            'album': 'album',
            'artist': 'artist',
            'duration': 'duration',
            'fanart': 'fanart',
            'file': 'file',
            'season': 'season',
            'showtitle': 'showtitle',
            'streamdetails': 'streamdetails',
            'thumbnail': 'thumbnail',
            'title': 'title',
            'tvshowid': 'tvshowid',
            'url': 'file',
        }

        app_props = {
            'volume': 'volume',
            'mute': 'muted',
        }

        player_props = {
            "duration": "totaltime",
            "position": "time",
            "repeat": "repeat",
            "seekable": "canseek",
            'speed': 'speed',
            "subtitles": "subtitles",
        }

        ret = {'state': PlayerState.IDLE.value}

        try:
            kodi = self._get_kodi()
            players = kodi.Player.GetActivePlayers().get('result', [])
        except:
            return ret

        ret['state'] = PlayerState.STOP.value
        app = kodi.Application.GetProperties(properties=list(set(app_props.values()))).get('result', {})

        for status_prop, kodi_prop in app_props.items():
            ret[status_prop] = app.get(kodi_prop)

        if not players:
            return ret

        if player_id is None:
            player_id = players.pop().get('playerid')
        else:
            for p in players:
                if p['player_id'] == player_id:
                    player_id = p
                    break

        if player_id is None:
            return ret

        media = kodi.Player.GetItem(playerid=player_id,
                                    properties=list(set(media_props.values()))).get('result', {}).get('item', {})

        for status_prop, kodi_prop in media_props.items():
            ret[status_prop] = media.get(kodi_prop)

        player_info = kodi.Player.GetProperties(
            playerid=player_id,
            properties=list(set(player_props.values()))).get('result', {})

        for status_prop, kodi_prop in player_props.items():
            ret[status_prop] = player_info.get(kodi_prop)

        if ret['duration']:
            ret['duration'] = self._time_obj_to_pos(ret['duration'])

        if ret['position']:
            ret['position'] = self._time_obj_to_pos(ret['position'])

        ret['state'] = PlayerState.PAUSE.value if player_info.get('speed', 0) == 0 else PlayerState.PLAY.value
        return ret


# vim:sw=4:ts=4:et:



from typing import List, Optional

from platypush.message.event import Event


class TelegramEvent(Event):
    def __init__(self, *args, chat_id: int, **kwargs):
        super().__init__(*args, chat_id=chat_id, **kwargs)


class MessageEvent(TelegramEvent):
    """
    Event triggered when a new message is received by the Telegram bot.
    """
    def __init__(self, *args, message, user, **kwargs):
        super().__init__(*args, message=message, user=user, **kwargs)


class CommandMessageEvent(MessageEvent):
    """
    Event triggered when a new message is received by the Telegram bot.
    """
    def __init__(self, command: str, cmdargs: Optional[List[str]] = None, *args, **kwargs):
        super().__init__(*args, command=command, cmdargs=(cmdargs or []), **kwargs)


class TextMessageEvent(MessageEvent):
    pass


class PhotoMessageEvent(MessageEvent):
    pass


class VideoMessageEvent(MessageEvent):
    pass


class ContactMessageEvent(MessageEvent):
    pass


class LocationMessageEvent(MessageEvent):
    pass


class DocumentMessageEvent(MessageEvent):
    pass


class GroupChatCreatedEvent(MessageEvent):
    pass


# vim:sw=4:ts=4:et:



from platypush.config import Config
from platypush.context import get_plugin
from platypush.plugins import Plugin, action


class VariablePlugin(Plugin):
    """
    This plugin allows you to manipulate context variables that can be
    accessed across your tasks. It requires the :mod:`platypush.plugins.db`
    and :mod:`platypush.plugins.redis` plugins to be enabled, as the variables
    will be stored either persisted on a local database or on the local Redis instance.

    Requires:

        * **sqlalchemy** (``pip install sqlalchemy``)
        * **redis** (``pip install redis``)
    """

    _variable_table_name = 'variable'

    def __init__(self, *args, **kwargs):
        """
        The plugin will create a table named ``variable`` on the database
        configured in the :mod:`platypush.plugins.db` plugin. You'll have
        to specify a default ``engine`` in your ``db`` plugin configuration.
        """

        super().__init__(*args, **kwargs)
        self.db_plugin = get_plugin('db')
        self.redis_plugin = get_plugin('redis')

        db = Config.get('db')
        self.db_config = {
            'engine': db.get('engine'),
            'args': db.get('args', []),
            'kwargs': db.get('kwargs', {})
        }

        self._create_tables()
        # self._variables = {}

    def _create_tables(self):
        self.db_plugin.execute("""CREATE TABLE IF NOT EXISTS {}(
            name varchar(255) not null primary key,
            value text
        )""".format(self._variable_table_name))

    @action
    def get(self, name, default_value=None):
        """
        Get the value of a variable by name from the local db.

        :param name: Variable name
        :type name: str

        :param default_value: What will be returned if the variable is not defined (default: None)

        :returns: A map in the format ``{"<name>":"<value>"}``
        """

        rows = self.db_plugin.select(table=self._variable_table_name,
                                     filter={'name': name},
                                     engine=self.db_config['engine'],
                                     *self.db_config['args'],
                                     **self.db_config['kwargs']).output

        return {name: rows[0]['value'] if rows else None}

    @action
    def set(self, **kwargs):
        """
        Set a variable or a set of variables on the local db.

        :param kwargs: Key-value list of variables to set (e.g. ``foo='bar', answer=42``)
        """

        records = [ { 'name': k, 'value': v }
                   for (k,v) in kwargs.items() ]

        self.db_plugin.insert(table=self._variable_table_name,
                              records=records, key_columns=['name'],
                              engine=self.db_config['engine'],
                              on_duplicate_update=True,
                              *self.db_config['args'],
                              **self.db_config['kwargs'])

        return kwargs


    @action
    def unset(self, name):
        """
        Unset a variable by name if it's set on the local db.

        :param name: Name of the variable to remove
        :type name: str
        """

        records = [ { 'name': name } ]

        self.db_plugin.delete(table=self._variable_table_name,
                              records=records, engine=self.db_config['engine'],
                              *self.db_config['args'],
                              **self.db_config['kwargs'])

        return True


    @action
    def mget(self, name):
        """
        Get the value of a variable by name from Redis.

        :param name: Variable name
        :type name: str

        :param default_value: What will be returned if the variable is not defined (default: None)

        :returns: A map in the format ``{"<name>":"<value>"}``
        """

        return self.redis_plugin.mget([name])

    @action
    def mset(self, **kwargs):
        """
        Set a variable or a set of variables on Redis.

        :param kwargs: Key-value list of variables to set (e.g. ``foo='bar', answer=42``)

        :returns: A map with the set variables
        """

        self.redis_plugin.mset(**kwargs)
        return kwargs


    @action
    def munset(self, name):
        """
        Unset a Redis variable by name if it's set

        :param name: Name of the variable to remove
        :type name: str
        """

        return self.redis_plugin.delete(name)


    @action
    def expire(self, name, expire):
        """
        Set a variable expiration on Redis

        :param name: Variable name
        :type name: str

        :param expire: Expiration time in seconds
        :type expire: int
        """

        return self.redis_plugin.expire(name, expire)


# vim:sw=4:ts=4:et:




import enum
import re
import textwrap

from xml.dom.minidom import parseString
import requests


class SwitchAction(enum.Enum):
    GET_STATE = 'GetBinaryState'
    SET_STATE = 'SetBinaryState'
    GET_NAME = 'GetFriendlyName'


class WemoRunner:
    default_port = 49153

    @staticmethod
    def exec(device: str, action: SwitchAction, port: int = default_port, value=None):
        state_name = action.value[3:]

        response = requests.post(
            'http://{}:{}/upnp/control/basicevent1'.format(device, port),
            headers={
                'User-Agent': '',
                'Accept': '',
                'Content-Type': 'text/xml; charset="utf-8"',
                'SOAPACTION': '\"urn:Belkin:service:basicevent:1#{}\"'.format(action.value),
            },
            data=re.sub('\s+', ' ', textwrap.dedent(
                '''
                <?xml version="1.0" encoding="utf-8"?>
                <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"
                        s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
                    <s:Body>
                        <u:{action} xmlns:u="urn:Belkin:service:basicevent:1">
                            <{state}>{value}</{state}>
                        </u:{action}
                    ></s:Body>
                </s:Envelope>
                '''.format(action=action.value, state=state_name,
                           value=value if value is not None else ''))))

        dom = parseString(response.text)
        return dom.getElementsByTagName(state_name).item(0).firstChild.data

    @classmethod
    def get_state(cls, device: str) -> bool:
        state = cls.exec(device=device, action=SwitchAction.GET_STATE)
        return bool(int(state))

    @classmethod
    def get_name(cls, device: str) -> str:
        name = cls.exec(device=device, action=SwitchAction.GET_NAME)
        return name

    @classmethod
    def on(cls, device):
        cls.exec(device=device, action=SwitchAction.SET_STATE, value=1)

    @classmethod
    def off(cls, device):
        cls.exec(device=device, action=SwitchAction.SET_STATE, value=0)

    @classmethod
    def toggle(cls, device):
        state = cls.get_state(device)
        cls.exec(device=device, action=SwitchAction.SET_STATE, value=int(not state))


# vim:sw=4:ts=4:et:



from redis import Redis

from platypush.context import get_backend
from platypush.plugins import Plugin, action


class RedisPlugin(Plugin):
    """
    Plugin to send messages on Redis queues.

    Requires:

        * **redis** (``pip install redis``)
    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs

        if not kwargs:
            # noinspection PyBroadException
            try:
                redis_backend = get_backend('redis')
                if redis_backend and redis_backend.redis_args:
                    self.kwargs = redis_backend.redis_args
            except:
                pass

    def _get_redis(self):
        return Redis(*self.args, **self.kwargs)

    @action
    def send_message(self, queue, msg, *args, **kwargs):
        """
        Send a message to a Redis queue.

        :param queue: Queue name
        :type queue: str

        :param msg: Message to be sent
        :type msg: str, bytes, list, dict, Message object

        :param args: Args passed to the Redis constructor (see https://redis-py.readthedocs.io/en/latest/#redis.Redis)
        :type args: list

        :param kwargs: Kwargs passed to the Redis constructor (see
            https://redis-py.readthedocs.io/en/latest/#redis.Redis)
        :type kwargs: dict
        """

        if args or kwargs:
            redis = Redis(*args, **kwargs)
        else:
            redis = self._get_redis()

        return redis.rpush(queue, str(msg))

    @action
    def mget(self, keys, *args):
        """
        :returns: The values specified in keys as a key/value dict (wraps MGET)
        """

        return {
            keys[i]: value.decode() if value else value
            for (i, value) in enumerate(self._get_redis().mget(keys, *args))
        }

    @action
    def mset(self, **kwargs):
        """
        Set key/values based on mapping (wraps MSET)
        """

        try:
            return self._get_redis().mset(**kwargs)
        except TypeError:
            # XXX commit https://github.com/andymccurdy/redis-py/commit/90a52dd5de111f0053bb3ebaa7c78f73a82a1e3e
            # broke back-compatibility with the previous way of passing
            # key-value pairs to mset directly on kwargs. This try-catch block
            # is to support things on all the redis-py versions
            return self._get_redis().mset(mapping=kwargs)

    @action
    def expire(self, key, expiration):
        """
        Set an expiration time in seconds for the specified key

        :param key: Key to set to expire
        :type key: str

        :param expiration: Expiration timeout (in seconds)
        :type expiration: int
        """

        return self._get_redis().expire(key, expiration)

    @action
    def delete(self, *args):
        """
        Delete one or multiple keys

        :param args: Keys to delete
        """

        return self._get_redis().delete(*args)


# vim:sw=4:ts=4:et:
